#!/usr/bin/env python
"""
XANES Normalization panel
"""
import time
import wx
from copy import deepcopy
import numpy as np

from functools import partial
from xraydb import guess_edge, atomic_number

from larch.utils import gformat, path_split
from larch.math import index_of
from larch.xafs.xafsutils import guess_energy_units
from larch.xafs.pre_edge import find_e0

from larch.wxlib import (BitmapButton, FloatCtrl, FloatSpin, get_icon,
                         SimpleText, pack, Button, HLine, Choice, Check,
                         GridPanel, CEN, RIGHT, LEFT, plotlabels,
                         get_zoomlimits, set_zoomlimits)

from larch.utils.strutils import fix_varname, fix_filename, file2groupname

from larch.utils.physical_constants import ATOM_NAMES
from larch.wxlib.plotter import last_cursor_pos
from .xas_dialogs import EnergyUnitsDialog
from .taskpanel import TaskPanel, autoset_fs_increment, update_confval
from .config import (make_array_choice, EDGES, ATSYMS,
                     NNORM_CHOICES, NNORM_STRINGS, NORM_METHODS)

np.seterr(all='ignore')

PLOTOPTS_1 = dict(style='solid', marker='None')
PLOTOPTS_2 = dict(style='short dashed', zorder=3, marker='None')
PLOTOPTS_D = dict(style='solid', zorder=2, side='right', marker='None')

PlotOne_Choices = make_array_choice(['mu','norm', 'flat', 'prelines',
                                     'norm+flat', 'mback_norm', 'mback_poly',
                                     'i0', 'norm+i0', 'dmude', 'norm+dmude',
                                     'd2mude', 'norm+d2mude'])


PlotSel_Choices = make_array_choice(['mu', 'norm', 'flat', 'dmude', 'd2mude'])

Plot_EnergyRanges = {'full E range': None,
                     'E0 -20:+80eV':  (-20, 80),
                     'E0 -30:+120eV': (-30, 120),
                     'E0 -50:+250eV': (-50, 250),
                     'E0 -100:+500eV': (-100, 500)}



FSIZE = 120
FSIZEBIG = 175

def get_auto_nnorm(norm1, norm2):
    "autoamatically set nnorm from range"
    nrange = abs(norm2 - norm1)
    nnorm = 2
    if nrange < 300:   nnorm = 1
    if nrange < 30:    nnorm = 0
    return nnorm

class XASNormPanel(TaskPanel):
    """XAS normalization Panel"""
    def __init__(self, parent, controller=None, **kws):
        TaskPanel.__init__(self, parent, controller, panel='xasnorm', **kws)

    def build_display(self):
        panel = self.panel
        self.wids = {}
        self.last_plot_type = 'one'

        trow = wx.Panel(panel)
        plot_sel = Button(trow, 'Plot Selected Groups', size=(175, -1),
                          action=self.onPlotSel)
        plot_one = Button(trow, 'Plot Current Group', size=(175, -1),
                          action=self.onPlotOne)

        self.plotsel_op = Choice(trow, choices=list(PlotSel_Choices.keys()),
                                 action=self.onPlotSel, size=(300, -1))
        self.plotone_op = Choice(trow, choices=list(PlotOne_Choices.keys()),
                                 action=self.onPlotOne, size=(300, -1))

        self.plot_erange = Choice(trow, choices=list(Plot_EnergyRanges.keys()),
                                 action=self.onPlotEither, size=(175, -1))

        opts = {'digits': 2, 'increment': 0.05, 'value': 0, 'size': (FSIZE, -1)}
        plot_voff = self.add_floatspin('plot_voff', with_pin=False,
                                       parent=trow,
                                       action=self.onVoffset,
                                       max_val=10000, min_val=-10000,
                                       **opts)

        vysize, vxsize = plot_sel.GetBestSize()
        voff_lab = wx.StaticText(parent=trow, label='  Y Offset:', size=(80, vxsize),
                                 style=wx.RIGHT|wx.ALIGN_CENTRE_HORIZONTAL|wx.ST_NO_AUTORESIZE)

        self.plot_erange.SetSelection(0)
        self.plotone_op.SetSelection(1)
        self.plotsel_op.SetSelection(1)

        tsizer = wx.GridBagSizer(3, 3)
        tsizer.Add(plot_sel,        (0, 0), (1, 1), LEFT, 2)
        tsizer.Add(self.plotsel_op, (0, 1), (1, 1), LEFT, 2)
        tsizer.Add(voff_lab,        (0, 2), (1, 1), RIGHT, 2)
        tsizer.Add(plot_voff,       (0, 3), (1, 1), RIGHT, 2)
        tsizer.Add(plot_one,       (1, 0), (1, 1), LEFT, 2)
        tsizer.Add(self.plotone_op, (1, 1), (1, 1), LEFT, 2)
        tsizer.Add(self.plot_erange, (1, 2), (1, 2), RIGHT, 2)

        pack(trow, tsizer)

        # atom row
        atpanel = wx.Panel(panel)
        self.wids['atsym']  = Choice(atpanel, choices=ATSYMS, size=(100, -1))
        self.wids['edge']   = Choice(atpanel, choices=EDGES, size=(100, -1))
        sat = wx.BoxSizer(wx.HORIZONTAL)
        sat.Add(self.wids['atsym'], 0, LEFT, 4)
        sat.Add(self.wids['edge'], 0, LEFT, 4)
        pack(atpanel, sat)

        # e0 row
        e0_panel = wx.Panel(panel)
        xas_e0   = self.add_floatspin('e0', action=self.onSet_XASE0Val,
                                      min_val=-1000, max_val=1e7, digits=3,
                                      increment=0.05, value=0, size=(FSIZEBIG, -1),
                                      parent=e0_panel)

        self.wids['auto_e0'] = Check(e0_panel, default=True, label='auto?',
                                    action=self.onAuto_XASE0)
        self.wids['show_e0'] = Check(e0_panel, default=True, label='show?',
                                     action=self.onSet_XASE0)

        sx = wx.BoxSizer(wx.HORIZONTAL)
        sx.Add(xas_e0, 0, LEFT, 4)
        sx.Add(self.wids['auto_e0'], 0, LEFT, 4)
        sx.Add(self.wids['show_e0'], 0, LEFT, 4)
        pack(e0_panel, sx)

        # step row
        step_panel = wx.Panel(panel)
        xas_step = self.add_floatspin('step', action=self.onSet_XASStep, with_pin=False,
                                      min_val=-1000.0, max_val=1e7, digits=4, increment=0.05,
                                      value=0.1, size=(FSIZEBIG, -1), parent=step_panel)
        self.wids['auto_step'] = Check(step_panel, default=True, label='auto?',
                                       action=self.onNormMethod)
        sx = wx.BoxSizer(wx.HORIZONTAL)
        sx.Add(xas_step, 0, LEFT, 4)
        sx.Add(self.wids['auto_step'], 0, LEFT, 4)
        pack(step_panel, sx)

        # step rows
        nnorm_panel = wx.Panel(panel)
        self.wids['nnorm'] = Choice(nnorm_panel, choices=list(NNORM_CHOICES.keys()),
                                    size=(150, -1), action=self.onNNormChoice,
                                    default=2)
        self.wids['auto_nnorm'] = Check(nnorm_panel, default=True, label='auto?',
                                    action=self.onAuto_NNORM)

        sx = wx.BoxSizer(wx.HORIZONTAL)
        sx.Add(self.wids['nnorm'], 0, LEFT, 4)
        sx.Add(self.wids['auto_nnorm'], 0, LEFT, 4)
        pack(nnorm_panel, sx)

        self.wids['energy_ref'] = Choice(panel, choices=['None'],
                                         action=self.onEnergyRef, size=(300, -1))

        self.wids['nvict'] = Choice(panel, choices=('0', '1', '2', '3'),
                                    size=(100, -1), action=self.onNormMethod,
                                    default=0)


        opts = {'size': (FSIZE, -1), 'digits': 2, 'increment': 5.0, 'action': self.onSet_Ranges,
                'min_val':-99000, 'max_val':99000}
        defaults = self.get_defaultconfig()

        pre_panel = wx.Panel(panel)
        xas_pre1 = self.add_floatspin('pre1', value=defaults['pre1'], parent=pre_panel, **opts)
        xas_pre2 = self.add_floatspin('pre2', value=defaults['pre2'], parent=pre_panel, **opts)
        self.wids['show_pre'] = Check(pre_panel, default=False, label='show?',
                                      action=self.onSet_XASE0)

        sx = wx.BoxSizer(wx.HORIZONTAL)
        sx.Add(xas_pre1, 0, LEFT, 4)
        sx.Add(SimpleText(pre_panel, ' : '), 0, LEFT, 4)
        sx.Add(xas_pre2, 0, LEFT, 4)
        sx.Add(self.wids['show_pre'], 0, LEFT, 4)
        pack(pre_panel, sx)

        nor_panel = wx.Panel(panel)
        xas_norm1 = self.add_floatspin('norm1', value=defaults['norm1'], parent=nor_panel, **opts)
        xas_norm2 = self.add_floatspin('norm2', value=defaults['norm2'], parent=nor_panel, **opts)

        self.wids['show_norm'] = Check(nor_panel, default=False, label='show?',
                                       action=self.onSet_XASE0)
        sx = wx.BoxSizer(wx.HORIZONTAL)
        sx.Add(xas_norm1, 0, LEFT, 4)
        sx.Add(SimpleText(nor_panel, ' : '), 0, LEFT, 4)
        sx.Add(xas_norm2, 0, LEFT, 4)
        sx.Add(self.wids['show_norm'], 0, LEFT, 4)
        pack(nor_panel, sx)


        opts = {'digits': 3, 'increment': 0.05, 'value': 0, 'size': (FSIZEBIG, -1)}

        opts.update({'value': 1.0, 'digits': 3})

        self.wids['norm_method'] = Choice(panel, choices=NORM_METHODS,
                                          size=(150, -1), action=self.onNormMethod)
        self.wids['norm_method'].SetSelection(0)
        self.wids['energy_shift'] = FloatSpin(panel, value=0, digits=3, increment=0.05,
                                              min_val=-99000, max_val=99000,
                                              action=self.onSet_EnergyShift,
                                              size=(FSIZEBIG, -1))


        self.wids['is_frozen'] = Check(panel, default=False, label='Freeze Group',
                                       action=self.onFreezeGroup)

        use_auto = Button(panel, 'Use Default Settings', size=(200, -1),
                          action=self.onResetNorm)

        def CopyBtn(name):
            return Button(panel, 'Copy', size=(60, -1),
                          action=partial(self.onCopyParam, name))

        copy_all = Button(panel, 'Copy All Parameters', size=(200, -1),
                          action=partial(self.onCopyParam, 'all'))

        add_text = self.add_text
        HLINEWID = 700
        panel.Add(SimpleText(panel, 'XAS Pre-edge subtraction and Normalization',
                             size=(650, -1), **self.titleopts), style=LEFT, dcol=4)
        panel.Add(trow, dcol=4, newrow=True)

        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)
        add_text('XAS Data:')
        panel.Add(use_auto, dcol=1)
        panel.Add(SimpleText(panel, 'Copy to Selected Groups:'), style=RIGHT, dcol=2)

        add_text('Element and Edge: ', newrow=True)
        panel.Add(atpanel, dcol=2)
        panel.Add(CopyBtn('atsym'), dcol=1, style=RIGHT)

        add_text('Energy Reference : ')
        panel.Add(self.wids['energy_ref'], dcol=2)
        panel.Add(CopyBtn('energy_ref'), dcol=1, style=RIGHT)

        add_text('Energy Shift : ')
        panel.Add(self.wids['energy_shift'], dcol=2)
        panel.Add(CopyBtn('energy_shift'), dcol=1, style=RIGHT)

        add_text('E0 : ')
        panel.Add(e0_panel, dcol=2)
        panel.Add(CopyBtn('xas_e0'), dcol=1, style=RIGHT)

        add_text('Edge Step: ')
        panel.Add(step_panel, dcol=2)
        panel.Add(CopyBtn('xas_step'), dcol=1, style=RIGHT)

        panel.Add((5, 5), newrow=True)
        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)

        add_text('Pre-edge range: ')
        panel.Add(pre_panel, dcol=2)
        panel.Add(CopyBtn('xas_pre'), dcol=1, style=RIGHT)

        panel.Add(SimpleText(panel, 'Victoreen order:'), newrow=True)
        panel.Add(self.wids['nvict'], dcol=2)

        panel.Add((5, 5), newrow=True)
        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)

        add_text('Normalization : ')
        panel.Add(self.wids['norm_method'], dcol=2)
        panel.Add(CopyBtn('xas_norm'), dcol=1, style=RIGHT)

        add_text('Norm Energy range: ')
        panel.Add(nor_panel, dcol=2)
        panel.Add(SimpleText(panel, 'Polynomial Type:'), newrow=True)
        panel.Add(nnorm_panel, dcol=3)

        panel.Add(HLine(panel, size=(HLINEWID, 3)), dcol=4, newrow=True)
        panel.Add(self.wids['is_frozen'], newrow=True)
        panel.Add(copy_all, dcol=3, style=RIGHT)

        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((5, 5), 0, LEFT, 3)
        sizer.Add(panel, 0, LEFT, 3)
        sizer.Add((5, 5), 0, LEFT, 3)
        pack(self, sizer)

    def get_config(self, dgroup=None):
        """custom get_config to possibly inherit from Athena settings"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        if dgroup is None:
            return self.get_defaultconfig()
        self.read_form()


        defconf = self.get_defaultconfig()
        conf = getattr(dgroup.config, self.configname, defconf)

        for k, v in defconf.items():
            if k not in conf:
                conf[k] = v
        if conf.get('edge_step', None) is None:
            conf['edge_step'] = getattr(dgroup, 'edge_step', 1)

        atsym = '?'
        if hasattr(dgroup, 'element'):
            elem = getattr(dgroup, 'element', '?')
            try:
                z = int(elem)
                atsym = ATSYMS[z]
            except:
                pass
            if elem in ATSYMS[1:]:
                atsym = elem
            else:
                try:
                    if elem.lower() in ATOM_NAMES:
                        z = 1 + ATOM_NAMES.index(eleme.lower())
                        atsym = ATSYMS[z]
                except:
                    pass

        conf['atsym'] = atsym
        if atsym == '?':
            conf['atsym'] = getattr(dgroup, 'atsym', atsym)
        conf['edge'] = getattr(dgroup,'edge', conf['edge'])

        xeref = getattr(dgroup, 'energy_ref', '')
        fname = getattr(dgroup, 'filename', None)
        if fname is None:
            fname = getattr(dgroup, 'groupname', None)
            if fname is None:
                fname =file2groupname('unknown_group',
                                      symtable=self._larch.symtable)

        conf['energy_ref'] = getattr(dgroup, 'energy_ref', fname)

        if conf['energy_ref'] in (None, 'None'):
            conf['energy_ref'] = fname


        conf['energy_shift'] = getattr(dgroup,'energy_shift', conf['energy_shift'])

        if hasattr(dgroup, 'e0') and conf['atsym'] == '?':
            atsym, edge = guess_edge(dgroup.e0)
            conf['atsym'] = atsym
            conf['edge'] = edge


        if hasattr(dgroup, 'mback_params'):
            conf['atsym'] = getattr(dgroup.mback_params, 'atsym', conf['atsym'])
            conf['edge'] = getattr(dgroup.mback_params, 'edge', conf['edge'])

        setattr(dgroup.config, self.configname, conf)
        return conf

    def fill_form(self, dgroup):
        """fill in form from a data group"""
        opts = self.get_config(dgroup)
        self.skip_process = True
        if self.is_xasgroup(dgroup):
            if self.plotone_op.GetCount() != len(PlotOne_Choices.keys()):
                self.plotone_op.SetChoices(list(PlotOne_Choices.keys()))
                self.plotone_op.SetSelection(1)
            if self.plotsel_op.GetCount() != len(PlotSel_Choices.keys()):
                self.plotsel_op.SetChoices(list(PlotSel_Choices.keys()))
                self.plotsel_op.SetSelection(1)

            groupnames = list(self.controller.file_groups.keys())
            self.wids['energy_ref'].SetChoices(groupnames)
            eref = opts.get('energy_ref', 'no_energy_ref')

            for key, val in self.controller.file_groups.items():
                if eref in (val, key):
                    self.wids['energy_ref'].SetStringSelection(key)

            self.wids['e0'].SetValue(opts.get('e0', -1))
            edge_step = opts.get('edge_step', 1.0)

            if hasattr(dgroup, 'e0') and opts['atsym'] == '?':
                atsym, edge = guess_edge(dgroup.e0)
                opts['atsym'] = atsym
                opts['edge'] = edge

            self.wids['step'].SetValue(edge_step)
            autoset_fs_increment(self.wids['step'], edge_step)
            for attr in ('pre1', 'pre2', 'norm1', 'norm2'):
                val = opts.get(attr, None)
                if val is not None:
                    self.wids[attr].SetValue(val)
            self.set_nnorm_widget(opts.get('nnorm'))

            self.wids['energy_shift'].SetValue(opts['energy_shift'])
            self.wids['nvict'].SetStringSelection("%d" % opts['nvict'])
            self.wids['show_e0'].SetValue(opts['show_e0'])
            self.wids['auto_e0'].SetValue(opts['auto_e0'])
            self.wids['auto_nnorm'].SetValue(opts.get('auto_nnorm', 0))
            self.wids['auto_step'].SetValue(opts['auto_step'])

            self.wids['edge'].SetStringSelection(opts['edge'].title())
            self.wids['atsym'].SetStringSelection(opts['atsym'].title())
            self.wids['norm_method'].SetStringSelection(opts['norm_method'].lower())
            for attr in ('pre1', 'pre2', 'norm1', 'norm2', 'nnorm', 'edge',
                         'atsym', 'step', 'norm_method'):
                self.wids[attr].Enable()
            self.wids['show_pre'].SetValue(opts['show_pre'])
            self.wids['show_norm'].SetValue(opts['show_norm'])


        frozen = opts.get('is_frozen', False)
        frozen = getattr(dgroup, 'is_frozen', frozen)

        self.wids['is_frozen'].SetValue(frozen)
        self._set_frozen(frozen)
        wx.CallAfter(self.unset_skip_process)

    def set_nnorm_widget(self, nnorm=None):
        conf = self.get_config()
        nnorm_default = get_auto_nnorm(conf['norm1'], conf['norm2'])
        if nnorm in (None, 'auto'):
            nnorm = nnorm_default
        elif nnorm in NNORM_CHOICES:
            nnorm = int(NNORM_CHOICES[nnorm])
        nnorm_str = NNORM_STRINGS.get(nnorm, None)
        if nnorm_str is None:
            nnorm_str = NNORM_STRINGS.get(nnorm_default, '1')
        self.wids['nnorm'].SetStringSelection(nnorm_str)
        self.wids['auto_nnorm'].SetValue(0)

    def unset_skip_process(self):
        self.skip_process = False

    def read_form(self):
        "read form, return dict of values"
        form_opts = {}
        form_opts['e0'] = self.wids['e0'].GetValue()
        form_opts['edge_step'] = self.wids['step'].GetValue()
        for attr in ('pre1', 'pre2', 'norm1', 'norm2'):
            val = self.wids[attr].GetValue()
            if val == 0: val = None
            form_opts[attr] = val

        form_opts['energy_shift'] = self.wids['energy_shift'].GetValue()

        form_opts['nnorm'] = NNORM_CHOICES.get(self.wids['nnorm'].GetStringSelection(), 1)
        form_opts['nvict'] = int(self.wids['nvict'].GetStringSelection())
        form_opts['plotone_op'] = self.plotone_op.GetStringSelection()
        form_opts['plotsel_op'] = self.plotsel_op.GetStringSelection()
        form_opts['plot_voff'] = self.wids['plot_voff'].GetValue()
        for ch in ('show_e0', 'show_pre', 'show_norm', 'auto_e0',
                   'auto_step', 'auto_nnorm'):
            form_opts[ch] = self.wids[ch].IsChecked()

        form_opts['norm_method'] = self.wids['norm_method'].GetStringSelection().lower()
        form_opts['edge'] = self.wids['edge'].GetStringSelection().title()
        form_opts['atsym'] = self.wids['atsym'].GetStringSelection().title()
        form_opts['energy_ref'] = self.wids['energy_ref'].GetStringSelection()
        return form_opts

    def onNNormChoice(self, evt=None):
        auto_nnorm  = self.wids['auto_nnorm'].SetValue(0)
        self.onNormMethod()

    def onNormMethod(self, evt=None):
        method = self.wids['norm_method'].GetStringSelection().lower()
        auto_nnorm  = self.wids['auto_nnorm'].GetValue()

        nnorm  = NNORM_CHOICES.get(self.wids['nnorm'].GetStringSelection(), 1)
        if nnorm is None:
            nnorm = get_auto_nnorm(self.wids['norm1'].GetValue(),
                                   self.wids['norm2'].GetValue())

        nvict = int(self.wids['nvict'].GetStringSelection())
        self.update_config({'norm_method': method, 'nnorm': nnorm, 'nvict': nvict})
        if method.startswith('mback'):
            dgroup = self.controller.get_group()
            cur_elem = self.wids['atsym'].GetStringSelection()
            if hasattr(dgroup, 'e0') and cur_elem == 'H':
                atsym, edge = guess_edge(dgroup.e0)
                self.wids['edge'].SetStringSelection(edge)
                self.wids['atsym'].SetStringSelection(atsym)
                self.update_config({'edge': edge, 'atsym': atsym})
        time.sleep(0.002)
        wx.CallAfter(self.onReprocess)

    def _set_frozen(self, frozen):
        try:
            dgroup = self.controller.get_group()
            dgroup.is_frozen = frozen
        except:
            pass

        for wattr in ('e0', 'step', 'pre1', 'pre2', 'norm1', 'norm2', 'nvict',
                      'nnorm', 'show_e0', 'auto_e0', 'auto_step', 'auto_nnorm',
                      'norm_method', 'edge', 'atsym', 'show_pre', 'show_norm'):
            self.wids[wattr].Enable(not frozen)

    def onFreezeGroup(self, evt=None):
        self._set_frozen(evt.IsChecked())

    def onEnergyRef(self, evt=None):
        dgroup = self.controller.get_group()
        eref = self.wids['energy_ref'].GetStringSelection()
        gname = self.controller.file_groups[eref]
        dgroup.config.xasnorm['energy_ref'] = eref
        dgroup.energy_ref = eref
        self.update_config({'energy_ref': eref}, dgroup=dgroup)

    def onPlotEither(self, evt=None):
        if self.last_plot_type == 'multi':
            self.onPlotSel(evt=evt)
        else:
            self.onPlotOne(evt=evt)

    def onPlotOne(self, evt=None):
        self.last_plot_type = 'one'
        self.plot(self.controller.get_group())
        wx.CallAfter(self.controller.set_focus)

    def onVoffset(self, evt=None):
        time.sleep(0.002)
        wx.CallAfter(self.onPlotSel)

    def onPlotSel(self, evt=None):
        newplot = True
        self.last_plot_type = 'multi'
        group_ids = self.controller.filelist.GetCheckedStrings()
        if len(group_ids) < 1:
            return
        last_id = group_ids[-1]

        groupname = self.controller.file_groups[str(last_id)]
        dgroup = self.controller.get_group(groupname)

        plot_choices = PlotSel_Choices

        erange = Plot_EnergyRanges[self.plot_erange.GetStringSelection()]
        self.controller.set_plot_erange(erange)

        ytitle = self.plotsel_op.GetStringSelection()
        yarray_name = plot_choices.get(ytitle, 'norm')
        ylabel = getattr(plotlabels, yarray_name, ytitle)
        xlabel = getattr(dgroup, 'plot_xlabel', getattr(plotlabels, 'energy'))

        if yarray_name == 'norm':
            norm_method = self.wids['norm_method'].GetStringSelection().lower()
            if norm_method.startswith('mback'):
                yarray_name = 'norm_mback'
                ylabel = "%s (MBACK)" % ylabel
            elif  norm_method.startswith('area'):
                yarray_name = 'norm_area'
                ylabel = "%s (Area)" % ylabel

        voff = self.wids['plot_voff'].GetValue()
        plot_traces = []
        newplot = True
        plotopts = self.controller.get_plot_conf()
        popts = {'style': 'solid', 'marker': None}
        popts['linewidth'] = plotopts.pop('linewidth')
        popts['marksize'] = plotopts.pop('markersize')
        popts['grid'] = plotopts.pop('show_grid')
        popts['fullbox'] = plotopts.pop('show_fullbox')

        for ix, checked in enumerate(group_ids):
            groupname = self.controller.file_groups[str(checked)]
            dgroup = self.controller.get_group(groupname)
            if dgroup is None:
                continue
            self.ensure_xas_processed(dgroup, force_mback =('mback' in yarray_name))

            if erange is not None and hasattr(dgroup, 'e0') and 'xmin' not in popts:
                popts['xmin'] = dgroup.e0 + erange[0]
                popts['xmax'] = dgroup.e0 + erange[1]

            trace = {'xdata': dgroup.xplot,
                     'ydata': getattr(dgroup, yarray_name) + ix*voff,
                     'label': dgroup.filename, 'new': newplot}
            trace.update(popts)
            plot_traces.append(trace)
            newplot = False

        ppanel = self.controller.get_display(stacked=False).panel
        zoom_limits = get_zoomlimits(ppanel, dgroup)

        nplot_traces = len(ppanel.conf.traces)
        nplot_request = len(plot_traces)
        if nplot_request > nplot_traces:
            linecolors = ppanel.conf.linecolors
            ncols = len(linecolors)
            for i in range(nplot_traces, nplot_request+5):
                ppanel.conf.init_trace(i,  linecolors[i%ncols], 'dashed')

        #

        ppanel.plot_many(plot_traces, xlabel=plotlabels.energy, ylabel=ylabel,
                         zoom_limits=zoom_limits, show_legend=True)
        set_zoomlimits(ppanel, zoom_limits) or ppanel.unzoom_all()
        ppanel.canvas.draw()

        wx.CallAfter(self.controller.set_focus)

    def onAuto_NNORM(self, evt=None):
        if evt.IsChecked():
            nnorm = get_auto_nnorm(self.wids['norm1'].GetValue(),
                                   self.wids['norm2'].GetValue())
            self.set_nnorm_widget(nnorm)
            self.wids['auto_nnorm'].SetValue(0)
            time.sleep(0.001)
            wx.CallAfter(self.onReprocess)

    def onResetNorm(self, evt=None):
        auto_nnorm = self.wids['auto_nnorm'].GetValue()
        if auto_nnorm:
            nnorm = get_auto_nnorm(self.wids['norm1'].GetValue(),
                                   self.wids['norm2'].GetValue())
            self.set_nnorm_widget(nnorm)

        defaults = self.get_defaultconfig()

        self.wids['auto_step'].SetValue(1)
        self.wids['auto_e0'].SetValue(1)
        self.wids['auto_e0'].SetValue(1)
        self.wids['nvict'].SetSelection(0)
        for attr in ('pre1', 'pre2', 'norm1', 'norm2'):
            self.wids[attr].SetValue(defaults[attr])
        self.onReprocess()

    def onCopyAuto(self, evt=None):
        opts = dict(pre1=0, pre2=0, nvict=0, norm1=0, norm2=0,
                    norm_method='polynomial', nnorm=2, auto_e0=1,
                    auto_step=1)
        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            grp = self.controller.get_group(groupname)
            if grp != self.controller.group and not getattr(grp, 'is_frozen', False):
                self.update_config(opts, dgroup=grp)
                self.fill_form(grp)
                self.process(grp, force=True)


    def onSaveConfigBtn(self, evt=None):
        conf = self.get_config()
        conf.update(self.read_form())
        # self.set_defaultconfig(conf)

    def onCopyParam(self, name=None, evt=None):
        conf = self.get_config()
        form = self.read_form()
        conf.update(form)
        dgroup = self.controller.get_group()
        self.update_config(conf)
        self.fill_form(dgroup)
        opts = {}
        name = str(name)
        def copy_attrs(*args):
            for a in args:
                opts[a] = conf[a]
        if name == 'all':
            copy_attrs('e0', 'show_e0', 'auto_e0', 'edge_step',
                       'auto_step', 'energy_shift', 'pre1', 'pre2',
                       'nvict', 'atsym', 'edge', 'norm_method', 'nnorm',
                       'norm1', 'norm2', 'energy_ref')
        elif name == 'xas_e0':
            copy_attrs('e0', 'show_e0', 'auto_e0')
        elif name == 'xas_step':
            copy_attrs('edge_step', 'auto_step')
        elif name == 'energy_shift':
            copy_attrs('energy_shift')
        elif name == 'xas_pre':
            copy_attrs('pre1', 'pre2', 'nvict', 'show_pre')
        elif name == 'atsym':
            copy_attrs('atsym', 'edge')
        elif name == 'xas_norm':
            copy_attrs('norm_method', 'nnorm', 'norm1', 'norm2', 'show_norm')
        elif name == 'energy_ref':
            copy_attrs('energy_ref')
        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            grp = self.controller.get_group(groupname)
            if grp != self.controller.group and not getattr(grp, 'is_frozen', False):
                self.update_config(opts, dgroup=grp)
                for key, val in opts.items():
                    if hasattr(grp, key):
                        setattr(grp, key, val)
                self.fill_form(grp)
                self.process(grp, force=True)

    def onAuto_XASE0(self, evt=None):
        if evt.IsChecked():
            dgroup = self.controller.get_group()
            find_e0(dgroup)
            self.update_config({'e0': dgroup.e0})
            time.sleep(0.002)
            wx.CallAfter(self.onReprocess)

    def onSet_XASE0(self, evt=None, value=None):
        "handle setting auto e0 / show e0"
        auto_e0  = self.wids['auto_e0'].GetValue()
        self.update_config({'e0': self.wids['e0'].GetValue(),
                           'auto_e0':self.wids['auto_e0'].GetValue()})
        time.sleep(0.002)
        wx.CallAfter(self.onReprocess)

    def onSet_XASE0Val(self, evt=None, value=None):
        "handle setting e0"
        self.wids['auto_e0'].SetValue(0)
        self.update_config({'e0': self.wids['e0'].GetValue(),
                            'auto_e0':self.wids['auto_e0'].GetValue()})
        time.sleep(0.002)
        wx.CallAfter(self.onReprocess)

    def onSet_EnergyShift(self, evt=None, value=None):
        conf = self.get_config()
        if conf['auto_energy_shift']:
            eshift = self.wids['energy_shift'].GetValue()
            dgroup = self.controller.get_group()
            _eref = getattr(dgroup, 'energy_ref', '<;no eref;>')
            _gname = dgroup.groupname
            self.stale_groups = []
            for fname, gname in self.controller.file_groups.items():
                this = self.controller.get_group(gname)
                if _gname != gname and _eref == getattr(this, 'energy_ref', None):
                    this.energy_shift = this.config.xasnorm['energy_shift'] = eshift
                    self.stale_groups.append(this)

        wx.CallAfter(self.onReprocess)

    def onSet_XASStep(self, evt=None, value=None):
        "handle setting edge step"
        edge_step = self.wids['step'].GetValue()
        if edge_step < 0:
            self.wids['step'].SetValue(abs(edge_step))
        self.wids['auto_step'].SetValue(0)
        self.update_config({'edge_step': abs(edge_step), 'auto_step': False})
        autoset_fs_increment(self.wids['step'], abs(edge_step))
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)


    def onSet_Ranges(self, evt=None, **kws):
        conf = {}
        for attr in ('pre1', 'pre2', 'norm1', 'norm2'):
            conf[attr] = self.wids[attr].GetValue()
        self.update_config(conf)
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)

    def pin_callback(self, opt='__', xsel=None, relative_e0=True, **kws):
        """
        get last selected point from a specified plot window
        and fill in the value for the widget defined by `opt`.

        by default it finds the latest cursor position from the
        cursor history of the first 20 plot windows.
        """
        if xsel is None or opt not in self.wids:
            return

        e0 = self.wids['e0'].GetValue()
        if opt == 'e0':
            self.wids['e0'].SetValue(xsel)
            self.wids['auto_e0'].SetValue(0)
        elif opt in ('pre1', 'pre2', 'norm1', 'norm2'):
            self.wids[opt].SetValue(xsel-e0)
        time.sleep(0.01)
        wx.CallAfter(self.onReprocess)

    def onReprocess(self, evt=None, value=None, **kws):
        "handle request reprocess"
        if self.skip_process:
            return
        try:
            dgroup = self.controller.get_group()
        except TypeError:
            return
        if not hasattr(dgroup.config, self.configname):
            return
        form = self.read_form()
        self.process(dgroup=dgroup)
        if self.stale_groups is not None:
            for g in self.stale_groups:
                self.process(dgroup=g, force=True)
            self.stale_groups = None
        self.onPlotEither()


    def process(self, dgroup=None, force_mback=False, force=False, use_form=True, **kws):
        """ handle process (pre-edge/normalize) of XAS data from XAS form
        """
        if self.skip_process and not force:
            return
        if dgroup is None:
            dgroup = self.controller.get_group()
        if dgroup is None:
            return

        self.skip_process = True
        conf = self.get_config(dgroup)
        form = self.read_form()
        if not use_form:
            form.update(self.get_defaultconfig())

        form['group'] = dgroup.groupname
        groupnames = list(self.controller.file_groups.keys())
        self.wids['energy_ref'].SetChoices(groupnames)
        eref_sel = self.wids['energy_ref'].GetStringSelection()
        for key, val in self.controller.file_groups.items():
            if eref_sel in (val, key):
                self.wids['energy_ref'].SetStringSelection(key)


        en_units = getattr(dgroup, 'energy_units', None)
        if en_units is None:
            en_units = guess_energy_units(dgroup.energy)

        if en_units != 'eV':
            mono_dspace = getattr(dgroup, 'mono_dspace', 1)
            dlg = EnergyUnitsDialog(self.parent, dgroup.energy,
                                    unitname=en_units,
                                    dspace=mono_dspace)
            res = dlg.GetResponse()
            dlg.Destroy()
            if res.ok:
                en_units = res.units
                dgroup.mono_dspace = res.dspace
                dgroup.xplot = dgroup.energy = res.energy
        dgroup.energy_units = en_units

        if not hasattr(dgroup, 'e0'):
            e0 = find_e0(dgroup)
            if form['atsym'] == '?' and conf.get('atsym', '?') != '?':
                form['atsym'] = conf['atsym']
                form['edge'] = conf.get('edge', 'K')

        if form['atsym'] == '?':
            form['atsym'], form['edge'] = guess_edge(dgroup.e0)
        dgroup.atsym = form['atsym']
        dgroup.edge = form['edge']


        cmds = []
        # test whether the energy shift is 0 or is different from the current energy shift:
        ediff = 8.42e14  # just a huge energy step/shift
        eshift_current = getattr(dgroup, 'energy_shift', ediff)
        eshift = form.get('energy_shift', ediff)
        e1 = getattr(dgroup, 'energy', [ediff])
        e2 = getattr(dgroup, 'energy_orig', None)

        if (not isinstance(e2, np.ndarray) or (len(e1) != len(e2))):
            cmds.append("{group:s}.energy_orig = {group:s}.energy[:]")

        if (isinstance(e1, np.ndarray) and isinstance(e2, np.ndarray) and
            len(e1) == len(e2)):
            ediff = (e1-e2).min()

        if abs(eshift-ediff) > 1.e-5 or abs(eshift-eshift_current) > 1.e-5:
            if abs(eshift) > 1e15: eshift = 0.0
            cmds.extend(["{group:s}.energy_shift = {eshift:.4f}",
                         "{group:s}.energy = {group:s}.xplot = {group:s}.energy_orig + {group:s}.energy_shift"])

        if len(cmds) > 0:
            self.larch_eval(('\n'.join(cmds)).format(group=dgroup.groupname, eshift=eshift))

        e0 = form['e0']
        edge_step = form['edge_step']

        copts = [dgroup.groupname]
        if not form['auto_e0']:
            if e0 < max(dgroup.energy) and e0 > min(dgroup.energy):
                copts.append("e0=%.4f" % float(e0))

        if not form['auto_step']:
            copts.append("step=%s" % gformat(float(edge_step)))

        for attr in ('pre1', 'pre2', 'nvict', 'nnorm', 'norm1', 'norm2'):
            val = form[attr]
            if val is None or val == 'auto':
                val = 'None'
            elif attr in ('nvict', 'nnorm'):
                if val in NNORM_CHOICES:
                    val = NNORM_CHOICES[val]
                val = int(val)
            else:
                val = f"{float(val):.2f}"
            copts.append(f"{attr}={val}")
        # print("process PreEdge ", copts)
        self.larch_eval("pre_edge(%s)" % (', '.join(copts)))
        self.larch_eval("{group:s}.norm_poly = 1.0*{group:s}.norm".format(**form))
        if not hasattr(dgroup, 'e0'):
            self.skip_process = False
            dgroup.mu = dgroup.yplot * 1.0
            opts = {'group': dgroup.groupname}
            return

        norm_method = form['norm_method'].lower()
        form['normmeth'] = 'poly'

        dgroup.journal.add_ifnew('normalization_method', norm_method)

        if force_mback or norm_method.startswith('mback'):
            form['normmeth'] = 'mback'
            copts = [dgroup.groupname]
            copts.append("z=%d" % atomic_number(form['atsym']))
            copts.append("edge='%s'" % form['edge'])
            for attr in ('pre1', 'pre2', 'nvict', 'nnorm', 'norm1', 'norm2'):
                val = form[attr]
                if val is None or val == 'auto':
                    val = 'None'
                elif attr in ('nvict', 'nnorm'):
                    if val in NNORM_CHOICES:
                        val = NNORM_CHOICES[val]
                    val = int(val)
                else:
                    val = f"{float(val):.2f}"
                copts.append(f"{attr}={val}")
            self.larch_eval("mback_norm(%s)" % (', '.join(copts)))

            if form['auto_step']:
                norm_expr = """{group:s}.norm = 1.0*{group:s}.norm_{normmeth:s}
{group:s}.edge_step = 1.0*{group:s}.edge_step_{normmeth:s}"""
                self.larch_eval(norm_expr.format(**form))
            else:
                norm_expr = """{group:s}.norm = 1.0*{group:s}.norm_{normmeth:s}
{group:s}.norm *= {group:s}.edge_step_{normmeth:s}/{edge_step:.8f}"""
                self.larch_eval(norm_expr.format(**form))

        if norm_method.startswith('area'):
            form['normmeth'] = 'area'
            expr = """{group:s}.norm = 1.0*{group:s}.norm_{normmeth:s}
{group:s}.edge_step = 1.0*{group:s}.edge_step_{normmeth:s}"""
            self.larch_eval(expr.format(**form))


        if form['auto_e0'] and hasattr(dgroup, 'e0'):
            self.wids['e0'].SetValue(dgroup.e0)
        if form['auto_step'] and hasattr(dgroup, 'edge_step'):
            self.wids['step'].SetValue(dgroup.edge_step)
            autoset_fs_increment(self.wids['step'], dgroup.edge_step)

        if hasattr(dgroup, 'e0') and (conf.get('atsym', '?') == '?'):
            atsym, edge = guess_edge(dgroup.e0)
            conf['atsym'] = dgroup.atsym = atsym
            conf['edge'] = dgroup.edge = edge
        self.wids['atsym'].SetStringSelection(dgroup.atsym)
        self.wids['edge'].SetStringSelection(dgroup.edge)

        self.set_nnorm_widget(dgroup.pre_edge_details.nnorm)
        for attr in ('e0', 'edge_step'):
            conf[attr] = getattr(dgroup, attr)
        for attr in ('pre1', 'pre2', 'norm1', 'norm2'):
            conf[attr] = val = getattr(dgroup.pre_edge_details, attr, None)
            if val is not None:
                self.wids[attr].SetValue(val)

        if hasattr(dgroup, 'mback_params'): # from mback
            conf['atsym'] = getattr(dgroup.mback_params, 'atsym')
            conf['edge'] = getattr(dgroup.mback_params, 'edge')
        self.update_config(conf, dgroup=dgroup)
        # print("process updated conf  ", dgroup, conf)
        wx.CallAfter(self.unset_skip_process)


    def get_plot_arrays(self, dgroup):
        lab = plotlabels.norm
        if dgroup is None:
            return

        dgroup.plot_y2label = None
        dgroup.plot_xlabel = plotlabels.energy
        dgroup.plot_yarrays = [('norm', PLOTOPTS_1, lab)]

        req_attrs = ['e0', 'norm', 'dmude', 'd2mude', 'pre_edge']

        pchoice = PlotOne_Choices.get(self.plotone_op.GetStringSelection(), 'norm')

        if pchoice in ('mu', 'norm', 'i0', 'flat', 'dmude', 'd2mude'):
            lab = getattr(plotlabels, pchoice)
            dgroup.plot_yarrays = [(pchoice, PLOTOPTS_1, lab)]

        elif pchoice == 'prelines':
            lab = plotlabels.mu
            dgroup.plot_yarrays = [('mu', PLOTOPTS_1, plotlabels.mu),
                                   ('pre_edge', PLOTOPTS_2, 'pre edge'),
                                   ('post_edge', PLOTOPTS_2, 'post edge')]

        elif pchoice == 'norm+d2mude':
            lab = plotlabels.norm
            dgroup.plot_y2label = lab2 = plotlabels.d2mude
            dgroup.plot_yarrays = [('norm', PLOTOPTS_1, lab),
                                   ('d2mude', PLOTOPTS_D, lab2)]

        elif pchoice == 'norm+dmude':
            lab = plotlabels.norm
            dgroup.plot_y2label = lab2 = plotlabels.dmude
            dgroup.plot_yarrays = [('norm', PLOTOPTS_1, lab),
                                   ('dmude', PLOTOPTS_D, lab2)]
        elif pchoice == 'norm+i0':
            lab = plotlabels.norm
            dgroup.plot_y2label = lab2 = plotlabels.i0
            dgroup.plot_yarrays = [('norm', PLOTOPTS_1, lab),
                                   ('i0', PLOTOPTS_D, lab2)]
        elif pchoice == 'norm+flat':
            lab = plotlabels.norm
            dgroup.plot_y2label = lab2 = plotlabels.flat
            dgroup.plot_yarrays = [('norm', PLOTOPTS_1, lab),
                                   ('flat', PLOTOPTS_D, lab2)]
        elif pchoice == 'mback_norm':
            req_attrs.append('mback_norm')
            lab = plotlabels.mu
            if not hasattr(dgroup, 'mback_mu'):
                self.process(dgroup=dgroup, force_mback=True)
            dgroup.plot_yarrays = [('mu', PLOTOPTS_1, lab),
                                   ('mback_mu', PLOTOPTS_2, r'tabulated $\mu(E)$')]

        elif pchoice == 'mback_poly':
            req_attrs.append('mback_norm')
            lab = plotlabels.norm
            if not hasattr(dgroup, 'mback_mu'):
                self.process(dgroup=dgroup, force_mback=True)
            dgroup.plot_yarrays = [('norm_mback', PLOTOPTS_1, 'mback'),
                                   ('norm_poly', PLOTOPTS_2, 'polynomial')]

        elif pchoice == 'area_norm':
            dgroup.plot_yarrays = [('norm_area', PLOTOPTS_1, 'area'),
                                   ('norm_poly', PLOTOPTS_2, 'polynomial')]

        dgroup.plot_ylabel = lab
        needs_proc = False
        for attr in req_attrs:
            needs_proc = needs_proc or (not hasattr(dgroup, attr))

        if needs_proc:
            self.process(dgroup=dgroup, force=True)

        y4e0 = dgroup.yplot = getattr(dgroup, dgroup.plot_yarrays[0][0], dgroup.mu)
        dgroup.plot_extras = []

        popts = {'marker': 'o', 'markersize': 5,
                 'label': '_nolegend_',
                 'markerfacecolor': '#888',
                 'markeredgecolor': '#A00'}

        if self.wids['show_e0'].IsChecked():
            ie0 = index_of(dgroup.energy, dgroup.e0)
            dgroup.plot_extras.append(('marker', dgroup.e0, y4e0[ie0], popts))

        if self.wids['show_pre'].IsChecked() or self.wids['show_norm'].IsChecked():
            popts['markersize'] = 4
            wids = []
            if self.wids['show_pre'].IsChecked():  wids.extend(['pre1', 'pre2'])
            if self.wids['show_norm'].IsChecked():  wids.extend(['norm1', 'norm2'])
            for wid in wids:
                val = self.wids[wid].GetValue()
                ival = min(len(y4e0)-1, index_of(dgroup.energy, dgroup.e0 + val))
                dgroup.plot_extras.append(('marker', dgroup.e0+val, y4e0[ival], popts))

    def plot(self, dgroup, title=None, plot_yarrays=None, yoff=0,
             delay_draw=True, multi=False, new=True, with_extras=True, **kws):

        if self.skip_plotting:
            return
        ppanel = self.controller.get_display(stacked=False).panel

        plotcmd = ppanel.oplot
        if new:
            plotcmd = ppanel.plot

        erange = Plot_EnergyRanges[self.plot_erange.GetStringSelection()]
        self.controller.set_plot_erange(erange)

        groupname = getattr(dgroup, 'groupname', None)
        if groupname is None:
            return

        self.ensure_xas_processed(dgroup, force_mback=True)
        self.get_plot_arrays(dgroup)

        if plot_yarrays is None and hasattr(dgroup, 'plot_yarrays'):
            plot_yarrays = dgroup.plot_yarrays

        popts = self.controller.get_plot_conf()
        popts.update(kws)
        popts['grid'] = popts.pop('show_grid')
        popts['fullbox'] = popts.pop('show_fullbox')

        path, fname = path_split(dgroup.filename)
        if 'label' not in popts:
            popts['label'] = dgroup.plot_ylabel

        zoom_limits = get_zoomlimits(ppanel, dgroup)

        if erange is not None and hasattr(dgroup, 'e0'):
            popts['xmin'] = dgroup.e0 + erange[0]
            popts['xmax'] = dgroup.e0 + erange[1]

        popts['xlabel'] = dgroup.plot_xlabel
        popts['ylabel'] = dgroup.plot_ylabel
        if getattr(dgroup, 'plot_y2label', None) is not None:
            popts['y2label'] = dgroup.plot_y2label

        plot_choices = PlotSel_Choices
        if multi:
            ylabel = self.plotsel_op.GetStringSelection()
            yarray_name = plot_choices.get(ylabel, 'norm')

            if self.is_xasgroup(dgroup):
                ylabel = getattr(plotlabels, yarray_name, ylabel)
            popts['ylabel'] = ylabel

        plot_extras = None
        if new:
            if title is None:
                title = fname
            plot_extras = getattr(dgroup, 'plot_extras', None)

        popts['title'] = title
        popts['show_legend'] = len(plot_yarrays) > 1
        narr = len(plot_yarrays) - 1

        _linewidth = popts['linewidth']
        for i, pydat in enumerate(plot_yarrays):
            yaname, yopts, yalabel = pydat
            popts.update(yopts)
            if yalabel is not None:
                popts['label'] = yalabel
            linewidht = _linewidth
            if 'linewidth' in popts:
                linewidth = popts.pop('linewidth')
            popts['delay_draw'] = delay_draw
            if yaname == 'norm_mback' and not hasattr(dgroup, yaname):
                self.process(dgroup=dgroup, force=True, force_mback=True)
            if yaname == 'i0' and not hasattr(dgroup, yaname):
                dgroup.i0 = np.ones(len(dgroup.xplot))
            plotcmd(dgroup.xplot, getattr(dgroup, yaname)+yoff, linewidth=linewidth, **popts)
            plotcmd = ppanel.oplot

        if with_extras and plot_extras is not None:
            axes = ppanel.axes
            for etype, x, y, opts in plot_extras:
                if etype == 'marker':
                    xpopts = {'marker': 'o', 'markersize': 5,
                              'label': '_nolegend_',
                              'markerfacecolor': 'red',
                              'markeredgecolor': '#884444'}
                    xpopts.update(opts)
                    axes.plot([x], [y], **xpopts)
                elif etype == 'vline':
                    xpopts = {'ymin': 0, 'ymax': 1.0,
                              'label': '_nolegend_',
                              'color': '#888888'}
                    xpopts.update(opts)
                    axes.axvline(x, **xpopts)

        # set_zoomlimits(ppanel, zoom_limits)
        ppanel.reset_formats()
        set_zoomlimits(ppanel, zoom_limits)
        ppanel.conf.unzoom(full=True, delay_draw=False)



import json
import time
import numpy as np
import uuid, socket, platform
from collections import namedtuple

from gzip import GzipFile

from lmfit import Parameter, Parameters
# from lmfit.model import Model, ModelResult
# from lmfit.minimizer import Minimizer, MinimizerResult

from larch import Group, isgroup, __date__, __version__, __release_version__
from ..utils import (isotime, bytes2str, str2bytes, fix_varname,
                     read_textfile, unique_name, format_exception, unixpath)
from ..utils.jsonutils import encode4js, decode4js

SessionStore = namedtuple('SessionStore', ('config', 'command_history', 'symbols'))

EMPTY_FEFFCACHE = {'paths': {}, 'runs': {}}

def invert_dict(d):
    "invert a dictionary {k: v} -> {v: k}"
    return {v: k for k, v in d.items()}

def get_machineid():
    "machine id / MAC address, independent of hostname"
    return hex(uuid.getnode())[2:]

def is_larch_session_file(fname):
    return read_textfile(fname, size=64).startswith('##LARIX:')

def save_groups(fname, grouplist):
    """save a list of groups (and other supported datatypes) to file

    This is a simplified and minimal version of save_session()

    Use 'read_groups()' to read data saved from this function
    """
    buff = ["##LARCH GROUPLIST"]
    for dat in grouplist:
        buff.append(json.dumps(encode4js(dat)))

    buff.append("")

    fh = GzipFile(unixpath(fname), "w")
    fh.write(str2bytes("\n".join(buff)))
    fh.close()

def read_groups(fname):
    """read a list of groups (and other supported datatypes)
    from a file saved with 'save_groups()'

    Returns a list of objects
    """
    text = read_textfile(fname)
    lines = text.split('\n')
    line0 = lines.pop(0)
    if not line0.startswith('##LARCH GROUPLIST'):
        raise ValueError(f"Invalid Larch group file: '{fname:s}'")

    out = []
    for line in lines:
        if len(line) > 1:
            out.append(decode4js(json.loads(line)))
    return out


def save_session(fname=None, symbols=None, histbuff=None,
                auto_xasgroups=False, _larch=None):
    """save all groups and data into a Larch Save File (.larix)
    A portable compressed json file, that can be loaded with `read_session()`

    Arguments:
        fname (str):   name of output save file.
        symbols [list of str or None]: names of symbols to save. [None]
               saving all non-core (user-generated) objects.
        histbuff [list of str or None]: command history, [None]
               saving the full history of the current session.
        auto_xasgroups [bool]: whether to automatically generate the
               `_xasgroups` dictionary for "XAS Groups" as used by
               Larix, which will include all symbols that are Groups
               and have both 'filename' and 'groupname' attributes.

    Notes:
        1. if `symbols` is `None` (default), all variables outside of the
           core groups will be saved: this effectively saves "the whole session".
           A limited list of symbol names can also be provided, saving part of
           a project (say, one or two data sets).
        2. if `histbuff` is `None` (default), the full list of commands in
           the session will be saved.
        3. auto_xasgroups will generate an `_xasgroups` dictionary (used by
           Larix to decide what groups to present as "Data Groups") from the
           supplied or available symbols: every Group with a "groupname" and
           "filename" will be included.

    See Also:
        read_session, load_session, clear_sessio

    """
    if fname is None:
        fname = time.strftime('%Y%b%d_%H%M')
    if not fname.endswith('.larix'):
        fname = fname + '.larix'

    if _larch is None:
        raise ValueError('_larch not defined')
    symtab = _larch.symtable

    buff = ["##LARIX: 1.0      Larch Session File",
            "##Date Saved: %s"   % time.strftime('%Y-%m-%d %H:%M:%S'),
            "##<CONFIG>",
            "##Machine Platform: %s" % platform.system(),
            "##Machine Name: %s" % socket.gethostname(),
            "##Machine MACID: %s" % get_machineid(),
            "##Machine Version: %s"   % platform.version(),
            "##Machine Processor: %s" % platform.machine(),
            "##Machine Architecture: %s" % ':'.join(platform.architecture()),
            "##Python Version: %s" % platform.python_version(),
            "##Python Compiler: %s" % platform.python_compiler(),
            "##Python Implementation: %s" % platform.python_implementation(),
            "##Larch Release Version: %s" % __release_version__,
            "##Larch Release Date: %s" % __date__,
            ]

    core_groups = symtab._sys.core_groups
    buff.append('##Larch Core Groups: %s' % (json.dumps(core_groups)))

    config = symtab._sys.config
    for attr in dir(config):
        buff.append('##Larch %s: %s' % (attr, json.dumps(getattr(config, attr, None))))
    buff.append("##</CONFIG>")


    if histbuff is None:
        try:
            histbuff = _larch.input.history.get(session_only=True)
        except:
            histbuff = None

    if histbuff is not None:
        buff.append("##<Session Commands>")
        buff.extend(["%s" % l for l in histbuff])
        buff.append("##</Session Commands>")

    if symbols is None:
        symbols = []
        for attr in symtab.__dir__(): # insert order, not alphabetical order
            if attr not in core_groups:
                symbols.append(attr)
    nsyms = len(symbols)

    _xasgroups = None
    if '_xasgroups' not in symbols and auto_xasgroups:
        nsyms +=1
        _xasgroups = {}
        for sname in symbols:
            obj = getattr(symtab, sname, None)
            if isgroup(obj):
                gname = getattr(obj, 'groupname', None)
                fname = getattr(obj, 'filename', None)
                if gname is not None and fname is not None:
                    _xasgroups[fname] = gname

    buff.append("##<Symbols: count=%d>"  % len(symbols))
    if _xasgroups is not None:
        buff.append('<:_xasgroups:>')
        buff.append(json.dumps(encode4js(_xasgroups)))

    for attr in symbols:
        if attr not in core_groups:
            buff.append(f'<:{attr}:>')
            buff.append(json.dumps(encode4js(getattr(symtab, attr))))

    buff.append("##</Symbols>")
    buff.append("")

    fh = GzipFile(unixpath(fname), "w")
    fh.write(str2bytes("\n".join(buff)))
    fh.close()

def clear_session(_larch=None):
    """clear user-definded data in a session

    Example:
         >>> save_session('foo.larix')
         >>> clear_session()

     will effectively save and then reset the existing session.
    """
    if _larch is None:
        raise ValueError('_larch not defined')

    core_groups = _larch.symtable._sys.core_groups
    for attr in _larch.symtable.__dir__():
        if attr not in core_groups:
            delattr(_larch.symtable, attr)


def read_session(fname, clean_xasgroups=True):
    """read Larch Session File, returning data into new data in the
    current session

    Arguments:
         fname (str):  name of save file

    Returns:
       Tuple
       A tuple wih entries:

           | configuration  - a dict of configuration for the saved session.
           | command_history  - a list of commands in the saved session.
           | symbols         - a dict of Larch/Python symbols, groups, etc

    See Also:
       load_session


    """
    text = read_textfile(fname)
    lines = text.split('\n')
    line0 = lines.pop(0)
    if not line0.startswith('##LARIX:'):
        raise ValueError(f"Invalid Larch session file: '{fname:s}'")

    version = line0.split()[1]

    symbols = {}
    config = {'Larix Version': version}
    cmd_history = []
    nsyms = nsym_expected = 0
    section = symname = '_unknown_'
    for line in lines:
        if line.startswith("##<"):
            section = line.replace('##<','').replace('>', '').strip().lower()
            if ':' in section:
                section, options = section.split(':', 1)
            if section.startswith('/'):
                section = '_unknown_'
        elif section == 'session commands':
            cmd_history.append(line)

        elif section == 'symbols':
            if line.startswith('<:') and line.endswith(':>'):
                symname = line.replace('<:', '').replace(':>', '')
            else:
                try:
                    symbols[symname] = decode4js(json.loads(line))
                except:
                    print(''.join(format_exception()))
                    print("decode failed:: ", symname, repr(line)[:50])
        else:
            if line.startswith('##') and ':' in line:
                line = line[2:]
                key, val = line.split(':', 1)
                key = key.strip()
                val = val.strip()
                if '[' in val or '{' in val:
                    try:
                        val = decode4js(json.loads(val))
                    except:
                        print(''.join(format_exception()))
                        print("decode failed @## ", repr(val)[:50])
                config[key] = val
    if '_xasgroups' in symbols and clean_xasgroups:
        missing = []
        for name, group in symbols['_xasgroups'].items():
            if group not in symbols:
                missing.append(name)
        for name in missing:
            symbols['_xasgroups'].pop(name)

    return SessionStore(config, cmd_history, symbols)


def load_session(fname, ignore_groups=None, include_xasgroups=None, _larch=None, verbose=False):
    """load all data from a Larch Session File into current larch session,
    merging into existing groups as appropriate (see Notes below)

    Arguments:
       fname  (str):  name of session file
       ignore_groups (list of strings): list of symbols to not import
       include_xasgroups (list of strings): list of symbols to import as XAS spectra,
                           even if not expicitly set in `_xasgroups`
       verbose (bool): whether to print warnings for overwrites [False]
    Returns:
        None

    Notes:
        1. data in the following groups will be merged into existing session groups:
           `_feffpaths` : dict of "current feff paths"
           `_feffcache` : dict with cached feff paths and feff runs
           `_xasgroups` : dict mapping "File Name" and "Group Name", used in `Larix`

        2. to avoid name clashes, group and file names in the `_xasgroups` dictionary
           may be modified on loading

    """
    if _larch is None:
        raise ValueError('load session needs a larch session')

    session = read_session(fname)

    if ignore_groups is None:
        ignore_groups = []
    if include_xasgroups is None:
        include_xasgroups = []

    # special groups to merge into existing session:
    #  _feffpaths, _feffcache, _xasgroups
    s_symbols = session.symbols
    s_xasgroups = s_symbols.pop('_xasgroups', {})

    s_xasg_inv = invert_dict(s_xasgroups)

    s_feffpaths = s_symbols.pop('_feffpaths', {})
    s_feffcache = s_symbols.pop('_feffcache', EMPTY_FEFFCACHE)

    symtab = _larch.symtable
    if not hasattr(symtab, '_xasgroups'):
        symtab._xasgroups = {}
    if not hasattr(symtab, '_feffpaths'):
        symtab._feffpaths = {}
    if not hasattr(symtab, '_feffcache'):
        symtab._feffcache = EMPTY_FEFFCACHE

    if not hasattr(symtab._sys, 'restored_sessions'):
        symtab._sys.restored_sessions = {}
    restore_data = {'date': isotime(),
                    'config': session.config,
                    'command_history': session.command_history}
    symtab._sys.restored_sessions[fname] = restore_data

    c_xas_gnames = list(symtab._xasgroups.values())

    for sym, val in s_symbols.items():
        if sym in ignore_groups:
            if sym in s_xasgroups.values():
                s_key = s_xasg_inv[sym]
                s_xasgroups.pop(s_key)
                s_xasg_inv = invert_dict(s_xasgroups)

            continue
        if sym in c_xas_gnames or sym in include_xasgroups:
            newsym = unique_name(sym, c_xas_gnames)
            c_xas_gnames.append(newsym)
            if sym in s_xasgroups.values():
                s_key = s_xasg_inv[sym]
                s_xasgroups[s_key] = newsym
                s_xasg_inv = invert_dict(s_xasgroups)
            sym = newsym

        if verbose and hasattr(symtab, sym):
            print(f"warning overwriting '{sym}'")
        setattr(symtab, sym, val)

    symtab._feffpaths.update(s_feffpaths)

    symtab._xasgroups.update(s_xasgroups)
    missing = []
    for name, group in symtab._xasgroups.items():
        if group not in symtab:
            missing.append(name)
    for name in missing:
        symtab._xasgroups.pop(name)

    for name in ('paths', 'runs'):
        symtab._feffcache[name].update(s_feffcache[name])



#!/usr/bin/env python
""" Builtins for larch"""

import sys
import time

from . import utils
from .utils.show import _larch_builtins as show_builtins

from .larchlib import parse_group_args, Journal
from .symboltable import Group
from .version import show_version

from . import math
from . import io
from . import fitting
from . import xray
from . import xrf
from . import xafs
from . import xrd
from . import xrmmap
from . import wxlib

from .utils import physical_constants

__core_modules = [math, fitting, io, xray, xrf, xafs, xrd, xrmmap, wxlib]

try:
    from . import epics
    __core_modules.append(epics)
except ImportError:
    pass

if wxlib.HAS_WXPYTHON:
    try:
        from .wxlib import plotter
        from . import wxmap, wxxas, wxxrd
        __core_modules.extend([plotter, wxmap, wxxas, wxxrd])
    except:
        pass

# inherit most available symbols from python's __builtins__
from_builtin = [sym for sym in __builtins__ if not sym.startswith('__')]

# inherit these from math (many will be overridden by numpy)
from_math = ('acos', 'acosh', 'asin', 'asinh', 'atan', 'atan2', 'atanh',
            'ceil', 'copysign', 'cos', 'cosh', 'degrees', 'e', 'exp',
            'fabs', 'factorial', 'floor', 'fmod', 'frexp', 'fsum', 'hypot',
            'isinf', 'isnan', 'ldexp', 'log', 'log10', 'log1p', 'modf',
            'pi', 'pow', 'radians', 'sin', 'sinh', 'sqrt', 'tan', 'tanh',
            'trunc')

# inherit these from numpy
from_numpy = ('abs', 'add', 'all', 'amax', 'amin', 'angle', 'any', 'append',
    'arange', 'arccos', 'arccosh', 'arcsin', 'arcsinh', 'arctan', 'arctan2',
    'arctanh', 'argmax', 'argmin', 'argsort', 'argwhere', 'around', 'array',
    'asarray', 'atleast_1d', 'atleast_2d', 'atleast_3d', 'average', 'bartlett',
    'bitwise_and', 'bitwise_not', 'bitwise_or', 'bitwise_xor', 'blackman',
    'broadcast', 'ceil', 'choose', 'clip', 'column_stack', 'common_type',
    'complex128', 'compress', 'concatenate', 'conjugate', 'convolve',
    'copysign', 'corrcoef', 'correlate', 'cos', 'cosh', 'cov', 'cross',
    'cumprod', 'cumsum', 'datetime_data', 'deg2rad', 'degrees', 'delete',
    'diag', 'diag_indices', 'diag_indices_from', 'diagflat', 'diagonal',
    'diff', 'digitize', 'divide', 'dot', 'dsplit', 'dstack', 'dtype', 'e',
    'ediff1d', 'empty', 'empty_like', 'equal', 'exp', 'exp2', 'expand_dims',
    'expm1', 'extract', 'eye', 'fabs', 'fft', 'fill_diagonal', 'finfo', 'fix',
    'flatiter', 'flatnonzero', 'fliplr', 'flipud', 'float64', 'floor',
    'floor_divide', 'fmax', 'fmin', 'fmod', 'format_parser', 'frexp',
    'frombuffer', 'fromfile', 'fromfunction', 'fromiter', 'frompyfunc',
    'fromregex', 'fromstring', 'genfromtxt', 'getbufsize', 'geterr',
    'gradient', 'greater', 'greater_equal', 'hamming', 'hanning', 'histogram',
    'histogram2d', 'histogramdd', 'hsplit', 'hstack', 'hypot', 'i0',
    'identity', 'iinfo', 'imag', 'indices', 'inexact', 'inf', 'info', 'inner',
    'insert', 'int32', 'integer', 'interp', 'intersect1d', 'invert',
    'iscomplex', 'iscomplexobj', 'isfinite', 'isinf', 'isnan', 'isneginf',
    'isposinf', 'isreal', 'isrealobj', 'isscalar', 'iterable', 'kaiser',
    'kron', 'ldexp', 'left_shift', 'less', 'less_equal', 'linalg', 'linspace',
    'little_endian', 'loadtxt', 'log', 'log10', 'log1p', 'log2', 'logaddexp',
    'logaddexp2', 'logical_and', 'logical_not', 'logical_or', 'logical_xor',
    'logspace', 'longdouble', 'longlong', 'mask_indices', 'matrix', 'maximum',
    'may_share_memory', 'mean', 'median', 'memmap', 'meshgrid', 'minimum',
    'mintypecode', 'mod', 'modf', 'msort', 'multiply', 'nan', 'nan_to_num',
    'nanargmax', 'nanargmin', 'nanmax', 'nanmin', 'nansum', 'ndarray',
    'ndenumerate', 'ndim', 'ndindex', 'negative', 'nextafter', 'nonzero',
    'not_equal', 'number', 'ones', 'ones_like', 'outer', 'packbits',
    'percentile', 'pi', 'piecewise', 'place', 'poly', 'poly1d', 'polyadd',
    'polyder', 'polydiv', 'polyint', 'polymul', 'polynomial', 'polysub',
    'polyval', 'power', 'prod', 'ptp', 'put', 'putmask', 'rad2deg', 'radians',
    'random', 'ravel', 'real', 'real_if_close', 'reciprocal', 'record',
    'remainder', 'repeat', 'reshape', 'resize', 'right_shift', 'rint', 'roll',
    'rollaxis', 'roots', 'rot90', 'round', 'searchsorted', 'select',
    'setbufsize', 'setdiff1d', 'seterr', 'setxor1d', 'shape', 'short', 'sign',
    'signbit', 'signedinteger', 'sin', 'sinc', 'single', 'sinh', 'size',
    'sort', 'sort_complex', 'spacing', 'split', 'sqrt', 'square', 'squeeze',
    'std', 'subtract', 'sum', 'swapaxes', 'take', 'tan', 'tanh', 'tensordot',
    'tile', 'trace', 'transpose', 'tri', 'tril', 'tril_indices',
    'tril_indices_from', 'trim_zeros', 'triu', 'triu_indices',
    'triu_indices_from', 'true_divide', 'trunc', 'ubyte', 'uint', 'uint32',
    'union1d', 'unique', 'unravel_index', 'unsignedinteger', 'unwrap',
    'ushort', 'vander', 'var', 'vdot', 'vectorize', 'vsplit', 'vstack',
    'where', 'zeros', 'zeros_like')

numpy_renames = {'ln':'log', 'asin':'arcsin', 'acos':'arccos',
                 'atan':'arctan', 'atan2':'arctan2', 'atanh':'arctanh',
                 'acosh':'arccosh', 'asinh':'arcsinh', 'npy_save': 'save',
                 'npy_load': 'load', 'npy_copy': 'copy'}

constants = {}
for pconst_name in ('PLANCK_HC', 'AVOGADRO', 'AMU',
                    'R_ELECTRON_ANG','DEG2RAD', 'RAD2DEG'):
    constants[pconst_name] = getattr(physical_constants, pconst_name)



## More builtin commands, to set up the larch language:

# def _group(_larch=None, **kws):
#     """create a group"""
#     if _larch is None:
#         _larch = Group()
#     else:
#         group = _larch.symtable.create_group()
#     for key, val in kws.items():
#         setattr(group, key, val)
#     return group

def _eval(text, filename=None, _larch=None):
    """evaluate a string of larch text
    """
    if _larch is None:
        raise Warning("cannot eval string -- larch broken?")
    return _larch.eval(text, fname=filename)


def _run(filename=None, new_module=None, _larch=None):
    "execute the larch text in a file as larch code."
    if _larch is None:
        raise Warning(f"cannot run file '{filename:s}' -- larch broken?")
    return _larch.runfile(filename, new_module=new_module)

def _reload(mod, _larch=None):
    """reload a module, either larch or python"""
    if _larch is None:
        raise Warning(f"cannot reload module '{mod:s}' -- larch broken?")

    modname = None
    if mod in _larch.symtable._sys.modules.values():
        for k, v in _larch.symtable._sys.modules.items():
            if v == mod:
                modname = k
    elif mod in sys.modules.values():
        for k, v in sys.modules.items():
            if v == mod:
                modname = k
    elif (mod in _larch.symtable._sys.modules.keys() or
          mod in sys.modules.keys()):
        modname = mod

    if modname is not None:
        return _larch.import_module(modname, do_reload=True)

def _help(*args, _larch=None):
    "show help on topic or object"
    write = sys.stdout.write
    if _larch is not None:
        write = _larch.writer.write
    buff = []
    for arg in args:
        if _larch is not None and isinstance(arg, str):
            arg= _larch.symtable.get_symbol(arg, create=False)
        buff.append(repr(arg))
        if callable(arg) and arg.__doc__ is not None:
            buff.append(arg.__doc__)
    buff.append('')
    write('\n'.join(buff))


def _journal(*args, **kws):
    return Journal(*args, **kws)

def _dir(obj=None, _larch=None):
    "return directory of an object -- thin wrapper about python builtin"
    if obj is None and _larch is not None:
        obj = _larch.symtable
    return dir(obj)

def _subgroups(obj):
    "return list of subgroups"
    if isinstance(obj, Group):
        return obj._subgroups()
    raise Warning("subgroups() argument must be a group")

def _groupitems(obj):
    "returns group items as if items() method of a dict"
    if isinstance(obj, Group):
        return obj._members().items()
    raise Warning("group_items() argument must be a group")

def _which(sym, _larch=None):
    "return full path of object, or None if object cannot be found"
    if _larch is None:
        raise Warning("cannot run which() -- larch broken?")
    stable = _larch.symtable
    if hasattr(sym, '__name__'):
        sym = sym.__name__
    if isinstance(sym, str) and stable.has_symbol(sym):
        obj = stable.get_symbol(sym)
        if obj is not None:
            return '%s.%s' % (stable.get_parentpath(sym), sym)
    return None

def _exists(sym, _larch=None):
    "return True if a named symbol exists and can be found, False otherwise"
    return _which(sym, _larch=_larch) is not None

def _isgroup(obj, _larch=None):
    """return whether argument is a group or the name of a group

    With additional arguments (all must be strings), it also tests
    that the group has an an attribute named for each argument. This
    can be used to test not only if a object is a Group, but whether
    it a group with expected arguments.

        > x = 10
        > g = group(x=x, y=2)
        > isgroup(g), isgroup(x)
        True, False
        > isgroup('g'), isgroup('x')
        True, False
        > isgroup(g, 'x', 'y')
        True
        > isgroup(g, 'x', 'y', 'z')
        False

    """
    if (_larch is not None and
        isinstance(obj, str) and
        _larch.symtable.has_symbol(obj)):
        obj = _larch.symtable.get_symbol(obj)
    return isinstance(obj, Group)


def _pause(msg='Hit return to continue', _larch=None):
    if _larch is None:
        raise Warning("cannot pause() -- larch broken?")
    return input(msg)

def _sleep(t=0):
    return time.sleep(t)
_sleep.__doc__ = time.sleep.__doc__

def _time():
    return time.time()
_time.__doc__ = time.time.__doc__

def _strftime(format, *args):
    return time.strftime(format, *args)
_strftime.__doc__ = time.strftime.__doc__


def save_history(filename, session_only=False, maxlines=5000, _larch=None):
    """save history of larch commands to a file"""
    _larch.input.history.save(filename, session_only=session_only, maxlines=maxlines)

def show_history(max_lines=10000, _larch=None):
    """show history of larch commands"""
    nhist = min(max_lines, len(_larch.history.buffer))
    for hline in _larch.history.buffer[-nhist:]:
        _larch.writer.write("%s\n" % hline)

def init_display_group(_larch):
    symtab = _larch.symtable
    if not symtab.has_group('_sys.display'):
        symtab.new_group('_sys.display')
        colors = {}
        colors['text'] = {'color': None}
        colors['text2'] = {'color': 'cyan'}
        colors['comment'] = {'color': 'green'}
        colors['error'] = {'color': 'red',  'attrs': ['bold']}
        display = symtab._sys.display
        display.colors = colors
        display.use_color = True
        display.terminal = 'xterm'


_main_builtins = dict(group=Group, Group=Group, dir=_dir, which=_which,
                      exists=_exists, isgroup=_isgroup,
                      subgroups=_subgroups, group_items=_groupitems,
                      parse_group_args=parse_group_args, pause=_pause,
                      sleep=_sleep, systime=_time, clock=_time,
                      strftime=_strftime, reload=_reload, run=_run,
                      eval=_eval, help=_help, journal=_journal,
                      show_version=show_version,
                      save_history=save_history,
                      show_history=show_history)

_main_builtins.update(utils._larch_builtins)
_main_builtins.update(show_builtins)


# names to fill in the larch namespace at startup
init_builtins = dict(_builtin=_main_builtins)

# functions to run (with signature fcn(_larch)) at interpreter startup
init_funcs = [init_display_group]

# group/classes to register for save-restore
init_moddocs = {}

for cmod in __core_modules:
    if cmod is None:
        continue
    cmodname  = getattr(cmod, '_larch_name', cmod.__name__)
    if cmodname.startswith('larch.'):
        cmodname = cmodname.replace('larch.', '_')

    doc = getattr(cmod, '__DOC__', None)
    if doc is not None:
        init_moddocs[cmodname] = doc
    builtins = getattr(cmod, '_larch_builtins', {})
    init_fcn = getattr(cmod, '_larch_init', None)

    for bkey, bval in builtins.items():
        if bkey not in init_builtins:
            init_builtins[bkey] = bval
        else:
            init_builtins[bkey].update(bval)

    if init_fcn is not None:
        init_funcs.append(init_fcn)

# list of supported valid commands -- don't need parentheses for these
valid_commands = ['run', 'help', 'show', 'which', 'more', 'cd']



#!/usr/bin/env python
"""

"""
import re
from copy import deepcopy

import numpy as np
np.seterr(all='ignore')
from pathlib import Path
from functools import partial

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.agw.flatnotebook as fnb
from wxmplot import PlotPanel

from wxutils import (SimpleText, FloatCtrl, FloatSpin, GUIColors, Button, Choice,
                     TextCtrl, pack, Popup, Check, MenuItem, CEN, RIGHT, LEFT,
                     FRAMESTYLE, HLine, Font)

import larch
from larch import Group
from larch.xafs.xafsutils import guess_energy_units
from larch.utils.strutils import fix_varname, fix_filename, file2groupname
from larch.io import look_for_nans,  guess_filereader, is_specfile, sum_fluor_channels
from larch.utils.physical_constants import PLANCK_HC, DEG2RAD
from larch.utils import gformat
from larch.math import safe_log
from . import FONTSIZE

CEN |=  wx.ALL
FNB_STYLE = fnb.FNB_NO_X_BUTTON|fnb.FNB_SMART_TABS
FNB_STYLE |= fnb.FNB_NO_NAV_BUTTONS|fnb.FNB_NODRAG

YPRE_OPS = ('', 'log(', '-log(', '-')
ARR_OPS = ('+', '-', '*', '/')

YERR_OPS = ('Constant', 'Sqrt(Y)', 'Array')
CONV_OPS  = ('Lorenztian', 'Gaussian')

DATATYPES = ('xydata', 'xas')
ENUNITS_TYPES = ('eV', 'keV', 'degrees', 'not energy')


MULTICHANNEL_TITLE = """ Sum MultiChannel Fluorescence Data, with Dead-Time Corrections:
  To allow for many Dead-Time-Correction methods, each Channel is built as:
        ROI_Corrected = ROI * ICR /(OCR * LTIME)

  Set the Number of Channels, the Step (usually 1) between columns for
  ROI 1, 2, ..., NChans, and any Bad Channels: a list of Channel numbers (start at 1).

  Select columns for ROI (counts) and correction factors ICR, OCR, and LTIME for Channel 1.

"""

ROI_STEP_TOOLTIP =  """number of columns between ROI columns -- typically 1 if the columns are like
   ROI_Ch1 ROI_Ch2 ROI_Ch3 ... ICR_Ch1 ICR_Ch2 ICR_Ch3 ... OCR_Ch1 OCR_Ch2 OCR_Ch3 ...

but set to 3 if the columns are arranged as
   ROI_Ch1 ICR_Ch1 OCR_Ch1 ROI_Ch2 ICR_Ch2 OCR_Ch2 ROI_Ch3 ICR_Ch3 OCR_Ch3 ...
"""
MAXCHANS=2000

class DeadtimeCorrectionFrame(wx.Frame):
    """Manage MultiChannel Fluorescence Data"""
    def __init__(self, parent, group, config=None, on_ok=None):
        self.parent = parent
        self.group = group
        self.config = {'bad_chans': [], 'plot_chan': 1, 'nchans': 4, 'step': 1,
                       'roi': '1.0', 'icr': '1.0', 'ocr': '1.0',
                       'ltime': '1.0', 'i0': '1.0'}
                       # 'out_choice': 'summed spectrum',
        if config is not None:
            self.config.update(config)
        self.arrays = {}
        self.on_ok = on_ok
        wx.Frame.__init__(self, None, -1, 'MultiChannel Fluorescence Data',
                          style=wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL)

        self.SetFont(Font(FONTSIZE))
        sizer = wx.GridBagSizer(2, 2)
        panel = scrolled.ScrolledPanel(self)

        self.SetMinSize((650, 450))
        self.yarr_labels = [s for s in self.parent.yarr_labels]
        wids = self.wids = {}

        multi_title = wx.StaticText(panel, label=MULTICHANNEL_TITLE, size=(650, 150))
        multi_title.SetFont(Font(FONTSIZE-1))
        for s in ('roi', 'icr', 'ocr', 'ltime'):
            wids[s] = Choice(panel, choices=self.yarr_labels, action=self.read_form, size=(150, -1))
            sel = self.config.get(s, '1.0')
            if sel == '1.0':
                wids[s].SetStringSelection(sel)
            else:
                wids[s].SetSelection(sel[0])
            wids[f'{s}_txt'] = SimpleText(panel, label='<list of column labels>', size=(275, -1))

        wids['i0'] = Choice(panel, choices=self.yarr_labels, action=self.read_form, size=(150, -1))
        wids['i0'].SetToolTip("All Channels will be divided by the I0 array")

        wids['i0'].SetStringSelection(self.parent.yarr2.GetStringSelection())

        wids['nchans']  = FloatCtrl(panel, value=self.config.get('nchans', 4),
                                    precision=0, maxval=MAXCHANS, minval=1, size=(50, -1),
                                    action=self.on_nchans)
        wids['bad_chans'] = TextCtrl(panel, value='', size=(175, -1), action=self.read_form)
        bad_chans = self.config.get('bad_chans', [])
        if len(bad_chans) > 0:
            wids['bad_chans'].SetValue(', '.join(['%d' % c for c in bad_chans]))
        wids['bad_chans'].SetToolTip("List Channels to skip, separated by commas or spaces")
        wids['step']    = FloatCtrl(panel, value=self.config.get('step', 1), precision=0,
                                    maxval=MAXCHANS, minval=1, size=(50, -1), action=self.read_form)
        wids['step'].SetToolTip(ROI_STEP_TOOLTIP)

        wids['plot_chan'] = FloatSpin(panel, value=self.config.get('plot_chan', 1),
                                      digits=0, increment=1, max_val=MAXCHANS, min_val=1, size=(50, -1),
                                      action=self.onPlotThis)

        wids['plot_this'] = Button(panel, 'Plot ROI + Correction For Channel', action=self.onPlotThis)
        wids['plot_all'] = Button(panel, 'Plot All Channels', action=self.onPlotEach)
        wids['plot_sum'] = Button(panel, 'Plot Sum of Channels',   action=self.onPlotSum)
        wids['save_btn'] = Button(panel, 'Use this Sum of Channels',  action=self.onOK_DTC)

        def tlabel(t):
            return SimpleText(panel, label=t)

        sizer.Add(multi_title,       (0, 0), (2, 5), LEFT, 3)
        ir = 2
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 5), LEFT, 3)

        ir += 1
        sizer.Add(tlabel(' Number of Channels:'),   (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['nchans'],                   (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(tlabel(' Step between Channels:'), (ir, 2), (1, 1), LEFT, 3)
        sizer.Add(wids['step'],                     (ir, 3), (1, 1), LEFT, 3)

        ir += 1
        sizer.Add(tlabel(' Bad Channels :'),   (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['bad_chans'],           (ir, 1), (1, 2), LEFT, 3)

        ir += 1
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 5), LEFT, 3)

        ir += 1
        sizer.Add(tlabel(' Signal '),  (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(tlabel(' Array for Channel #1 '),  (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(tlabel(' Array Labels used for all Channels '),  (ir, 2), (1, 3), LEFT, 3)

        for s in ('roi', 'icr', 'ocr', 'ltime'):
            ir += 1
            sizer.Add(tlabel(f' {s.upper()} #1 : '), (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(wids[s],               (ir, 1), (1, 1), LEFT, 3)
            sizer.Add(wids[f'{s}_txt'],      (ir, 2), (1, 3), LEFT, 3)

        ir += 1
        sizer.Add(tlabel(' I0 : '),  (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['i0'],        (ir, 1), (1, 1), LEFT, 3)

        ir += 1
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 5), LEFT, 3)

        ir += 1
        sizer.Add(wids['plot_this'],   (ir, 0), (1, 2), LEFT, 3)
        sizer.Add(tlabel(' Channel:'), (ir, 2), (1, 1), LEFT, 3)
        sizer.Add(wids['plot_chan'],   (ir, 3), (1, 1), LEFT, 3)

        ir += 1
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 5), LEFT, 3)
        ir += 1
        sizer.Add(wids['plot_all'],  (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['plot_sum'],  (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(wids['save_btn'],  (ir, 2), (1, 2), LEFT, 3)

        pack(panel, sizer)
        panel.SetupScrolling()

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel, 1, wx.GROW|wx.ALL, 1)

        pack(self, mainsizer)
        self.Show()
        self.Raise()

    def get_en_i0(self):
        en = self.group.xplot
        i0 = 1.0
        if self.config['i0'] != '1.0':
            i0 = self.group.data[self.config['i0'], :]
        return en, i0

    def read_arrays(self, pchan):
        def get_array(name, pchan, default=1.0):
            out = default
            if self.config[name] != '1.0':
                ix = self.config[name][pchan-1]
                if ix > 0:
                    out = self.group.data[ix, :]
            return out
        roi = get_array('roi', pchan, default=None)
        icr = get_array('icr', pchan)
        ocr = get_array('ocr', pchan)
        ltime = get_array('ltime', pchan)
        return roi, icr*ocr/ltime

    def onOK_DTC(self, event=None):
        self.read_form()
        if callable(self.on_ok):
            self.on_ok(self.config)
        self.Destroy()

    def onPlotSum(self, event=None):
        self.read_form()
        en, i0 = self.get_en_i0()
        label, sum = sum_fluor_channels(self.group, self.config['roi'],
                                        icr=self.config['icr'],
                                        ocr=self.config['ocr'],
                                        ltime=self.config['ltime'],
                                        add_data=False)
        if sum is not None:
            popts = dict(marker=None, markersize=0, linewidth=2.5,
                         show_legend=True, ylabel=label, label=label,
                         xlabel='Energy (eV)')
            self.parent.plotpanel.plot(en, sum/i0, new=True, **popts)

    def onPlotEach(self, event=None):
        self.read_form()
        new = True
        en, i0 = self.get_en_i0()
        popts = dict(marker=None, markersize=0, linewidth=2.5,
                     show_legend=True,  xlabel='Energy (eV)',
                     ylabel=f'Corrected Channels')

        nused = 0
        for pchan in range(1, self.config['nchans']+1):
            roi, dtc = self.read_arrays(pchan)
            if roi is not None:
                popts['label'] = f'Chan{pchan} Corrected'
                if new:
                    self.parent.plotpanel.plot(en, roi*dtc/i0, new=True, **popts)
                    new = False
                else:
                    self.parent.plotpanel.oplot(en, roi*dtc/i0, **popts)

    def onPlotThis(self, event=None):
        self.read_form()
        en, i0 = self.get_en_i0()
        pchan = self.config['plot_chan']
        roi, dtc = self.read_arrays(pchan)
        if roi is None:
            return
        ylabel = self.wids['roi'].GetStringSelection()
        popts = dict(marker=None, markersize=0, linewidth=2.5, show_legend=True,
                     ylabel=f'Chan{pchan}', xlabel='Energy (eV)',
                     label=f'Chan{pchan} Raw')

        self.parent.plotpanel.plot(en, roi/i0, new=True, **popts)
        popts['label'] = f'Chan{pchan} Corrected'
        self.parent.plotpanel.oplot(en, roi*dtc/i0, **popts)

    def on_nchans(self, event=None, value=None, **kws):
        try:
            nchans = self.wids['nchans'].GetValue()
            pchan = self.wids['plot_chan'].GetValue()
            self.wids['plot_chan'].SetMax(nchans)
            self.wids['plot_chan'].SetValue(pchan)
        except:
            pass

    def read_form(self, event=None, value=None, **kws):
        try:
            wids = self.wids
            nchans = int(wids['nchans'].GetValue())
            step = int(wids['step'].GetValue())
            badchans = wids['bad_chans'].GetValue().replace(',', ' ').strip()
        except:
            return

        bad_channels = []
        if len(badchans) > 0:
            try:
                bad_channels = [int(s) for s in badchans.split()]
                wids['bad_chans'].SetBackgroundColour('#FFFFFF')
            except:
                bad_channels = []
                wids['bad_chans'].SetBackgroundColour('#F0B03080')

        pchan = int(wids['plot_chan'].GetValue())

        self.config['bad_chans'] = bad_channels
        self.config['plot_chan'] = pchan
        self.config['nchans'] = nchans
        self.config['step'] = step
        self.config['i0'] = wids['i0'].GetSelection()
        if wids['i0'].GetStringSelection() in ('1.0', ''):
            self.config['i0'] = '1.0'

        for s in ('roi', 'icr', 'ocr', 'ltime'):
            lab = wids[s].GetStringSelection()
            ilab = wids[s].GetSelection()
            if lab in ('1.0', ''):
                wids[f"{s}_txt"].SetLabel(lab)
                wids[f"{s}_txt"].SetToolTip(lab)
                self.config[s] = '1.0'
            else:
                chans = [ilab + i*step for i in range(nchans)]
                labs = []
                for i in range(nchans):
                    if (i+1) in bad_channels:
                        chans[i] = -1
                    else:
                        nchan = chans[i]
                        if nchan < len(self.group.array_labels):
                            labs.append(self.group.array_labels[nchan])
                wids[f"{s}_txt"].SetLabel(', '.join(labs))
                self.config[s] = chans


class MultiColumnFrame(wx.Frame) :
    """Select Multiple Columns for import, optional i0 channel"""
    def __init__(self, parent, group, config=None, on_ok=None):
        self.parent = parent
        self.group = group

        self.on_ok = on_ok
        wx.Frame.__init__(self, None, -1, 'Import Multiple Columns from a file',
                          style=wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL)

        self.config = {'channels': [], 'i0': '1.0'}
        if config is not None:
            self.config.update(config)

        self.SetFont(Font(FONTSIZE))
        sizer = wx.GridBagSizer(2, 2)
        panel = scrolled.ScrolledPanel(self)

        self.SetMinSize((475, 350))
        self.yarr_labels = [s for s in self.parent.yarr_labels]
        wids = self.wids = {}

        wids['i0'] = Choice(panel, choices=self.yarr_labels, size=(200, -1))
        wids['i0'].SetToolTip("All Channels will be divided by the I0 array")

        wids['i0'].SetStringSelection(self.parent.yarr2.GetStringSelection())

        bpanel = wx.Panel(panel)
        bsizer = wx.BoxSizer(wx.HORIZONTAL)

        bsizer.Add(Button(bpanel, ' Select All ', action=self.onSelAll))
        bsizer.Add(Button(bpanel, ' Select None ', action=self.onSelNone))
        bsizer.Add(Button(bpanel, ' Plot Selected ', action=self.onPlotSel))
        bsizer.Add(Button(bpanel, ' Import Selected Columns ', action=self.onOK_Multi))
        pack(bpanel, bsizer)

        ir = 0
        sizer.Add(bpanel,  (ir, 0), (1, 5), LEFT, 3)
        ir += 1
        sizer.Add(HLine(panel, size=(450, 2)),  (ir, 0), (1, 5), LEFT, 3)

        ir += 1
        sizer.Add(SimpleText(panel, label=' I0 Array: '), (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['i0'],                        (ir, 1), (1, 3), LEFT, 3)

        ir += 1
        sizer.Add(SimpleText(panel, label=' Array Name'),  (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(SimpleText(panel, label=' Select '),   (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(SimpleText(panel, label=' Plot'),       (ir, 2), (1, 1), LEFT, 3)

        array_labels = getattr(group, 'array_labels', self.yarr_labels)
        nlabels = len(array_labels)
        narrays, npts = group.data.shape
        for i in range(narrays):
            if i < nlabels:
                name = array_labels[i]
            else:
                name = f'unnamed_column{i+1}'
            self.wids[f'use_{i}'] = chuse = Check(panel, label='', default=(i in self.config['channels']))
            slabel = SimpleText(panel, label=f' {name}  ')
            bplot  = Button(panel, 'Plot', action=partial(self.onPlot, index=i))

            ir += 1
            sizer.Add(slabel, (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(chuse,  (ir, 1), (1, 1), LEFT, 3)
            sizer.Add(bplot,  (ir, 2), (1, 1), LEFT, 3)

        pack(panel, sizer)
        panel.SetupScrolling()

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel, 1, wx.GROW|wx.ALL, 1)

        pack(self, mainsizer)
        self.Show()
        self.Raise()


    def onSelAll(self, event=None, *kws):
        for wname, wid in self.wids.items():
            if wname.startswith('use_'):
                wid.SetValue(1)

    def onSelNone(self, event=None, *kws):
        for wname, wid in self.wids.items():
            if wname.startswith('use_'):
                wid.SetValue(0)

    def onPlotSel(self, event=None):
        group = self.group
        self.config['i0']  = self.wids['i0'].GetSelection()
        channels = []
        x = self.group.xplot
        popts = dict(marker=None, markersize=0, linewidth=2.5,
                     ylabel='selected arrays', show_legend=True,
                     xlabel=self.group.plot_xlabel, delay_draw=True)
        first = True
        for wname, wid in self.wids.items():
            if wname.startswith('use_') and wid.IsChecked():
                chan = int(wname.replace('use_', ''))
                y = self.group.data[chan, :]
                try:
                    label = self.group.array_labels[chan]
                except:
                    label = f'column {chan+1}'
                plot = self.parent.plotpanel.oplot
                if first:
                    first = False
                    plot = self.parent.plotpanel.plot
                plot(x, y, label=label, **popts)
        self.parent.plotpanel.draw()


    def onPlot(self, event=None, index=None):
        if index is not None:
            x = self.group.xplot
            y = self.group.data[index, :]
            try:
                label = self.group.array_labels[index]
            except:
                label = f'column {index+1}'

            popts = dict(marker=None, markersize=0, linewidth=2.5,
                         ylabel=label, xlabel=self.group.plot_xlabel, label=label)
            self.parent.plotpanel.plot(x, y, **popts)

    def onOK_Multi(self, evt=None):
        group = self.group
        self.config['i0']  = self.wids['i0'].GetSelection()
        channels = []
        for wname, wid in self.wids.items():
            if wname.startswith('use_') and wid.IsChecked():
                chan = int(wname.replace('use_', ''))
                channels.append(chan)

        self.config['channels']  = channels
        if callable(self.on_ok):
            self.on_ok(self.config)
        self.Destroy()


class EditColumnFrame(wx.Frame) :
    """Edit Column Labels for a larch grouop"""
    def __init__(self, parent, group, on_ok=None):
        self.parent = parent
        self.group = group
        self.on_ok = on_ok
        wx.Frame.__init__(self, None, -1, 'Edit Array Names',
                          style=wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL)

        self.SetFont(Font(FONTSIZE))
        sizer = wx.GridBagSizer(2, 2)
        panel = scrolled.ScrolledPanel(self)

        self.SetMinSize((700, 450))

        self.wids = {}
        ir = 0
        sizer.Add(Button(panel, 'Apply Changes', size=(200, -1),
                         action=self.onOK_Edit),
                  (0, 1), (1, 2), LEFT, 3)
        sizer.Add(Button(panel, 'Use Column Number', size=(200, -1),
                         action=self.onColNumber),
                  (0, 3), (1, 2), LEFT, 3)
        sizer.Add(HLine(panel, size=(550, 2)),
                  (1, 1), (1, 5), LEFT, 3)

        cind = SimpleText(panel, label='Column')
        cold = SimpleText(panel, label=' Current Name')
        cnew = SimpleText(panel, label=' Enter New Name')
        cret = SimpleText(panel, label=' Result   ')
        cinfo = SimpleText(panel, label=' Data Range')
        cplot = SimpleText(panel, label=' Plot')

        ir = 2
        sizer.Add(cind,  (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(cold,  (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(cnew,  (ir, 2), (1, 1), LEFT, 3)
        sizer.Add(cret,  (ir, 3), (1, 1), LEFT, 3)
        sizer.Add(cinfo, (ir, 4), (1, 1), LEFT, 3)
        sizer.Add(cplot, (ir, 5), (1, 1), LEFT, 3)

        nlabels = len(group.array_labels)
        narrays, npts = group.data.shape
        for i in range(narrays):
            if i < nlabels:
                name = group.array_labels[i]
            else:
                name = f'unnamed_column{i+1}'
            ir += 1
            cind = SimpleText(panel, label=f' {i+1} ')
            cold = SimpleText(panel, label=f' {name} ')
            cret = SimpleText(panel, label=fix_varname(name))
            cnew = wx.TextCtrl(panel, value=name, size=(150, -1),
                               style=wx.TE_PROCESS_ENTER)

            cnew.Bind(wx.EVT_TEXT_ENTER, partial(self.update, index=i))
            cnew.Bind(wx.EVT_KILL_FOCUS, partial(self.update, index=i))

            arr = group.data[i,:]
            info_str = f" [{gformat(arr.min(),length=9)}:{gformat(arr.max(), length=9)}] "
            cinfo = SimpleText(panel, label=info_str)
            cplot = Button(panel, 'Plot', action=partial(self.onPlot, index=i))


            self.wids[f"{i}"] = cnew
            self.wids[f"ret_{i}"] = cret

            sizer.Add(cind,  (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(cold,  (ir, 1), (1, 1), LEFT, 3)
            sizer.Add(cnew,  (ir, 2), (1, 1), LEFT, 3)
            sizer.Add(cret,  (ir, 3), (1, 1), LEFT, 3)
            sizer.Add(cinfo, (ir, 4), (1, 1), LEFT, 3)
            sizer.Add(cplot, (ir, 5), (1, 1), LEFT, 3)

        pack(panel, sizer)
        panel.SetupScrolling()

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel, 1, wx.GROW|wx.ALL, 1)

        pack(self, mainsizer)
        self.Show()
        self.Raise()

    def onPlot(self, event=None, index=None):
        if index is not None:
            x = self.group.index
            y = self.group.data[index, :]
            label = self.wids["ret_%i" % index].GetLabel()
            popts = dict(marker='o', markersize=4, linewidth=1.5,
                         ylabel=label, xlabel='data point', label=label)
            self.parent.plotpanel.plot(x, y, **popts)

    def onColNumber(self, evt=None, index=-1):
        for name, wid in self.wids.items():
            val = name
            if name.startswith('ret_'):
                val = name[4:]
                setter = wid.SetLabel
            else:
                setter = wid.SetValue
            setter("col%d" % (int(val) +1))

    def update(self, evt=None, index=-1):
        newval = fix_varname(self.wids[f"{index}"].GetValue())
        self.wids[f"ret_{index}"].SetLabel(newval)

    def update_char(self, evt=None, index=-1):
        if evt.GetKeyCode() == wx.WXK_RETURN:
            self.update(evt=evt, index=index)
        # evt.Skip()

    def onOK_Edit(self, evt=None):
        group = self.group
        array_labels = []
        for i in range(len(self.group.array_labels)):
            newname = self.wids["ret_%i" % i].GetLabel()
            array_labels.append(newname)

        if callable(self.on_ok):
            self.on_ok(array_labels)
        self.Destroy()

class ColumnDataFileFrame(wx.Frame) :
    """Column Data File, select columns"""
    def __init__(self, parent, filename=None, groupname=None, config=None,
                 read_ok_cb=None, edit_groupname=True, _larch=None):
        self.parent = parent
        self._larch = _larch
        self.path = filename

        group = self.read_column_file(self.path)
        # print("COLUMN FILE Read ", self.path, getattr(group, 'datatype', 'unknown'))
        self.subframes = {}
        self.workgroup  = Group(raw=group)
        for attr in ('path', 'filename', 'groupname', 'datatype',
                     'array_labels', 'data'):
            setattr(self.workgroup, attr, getattr(group, attr, None))

        self.array_labels = [l.lower() for l in group.array_labels]

        has_energy = False
        en_units = 'unknown'
        for arrlab in self.array_labels[:5]:
            arrlab  = arrlab.lower()
            if arrlab.startswith('en') or 'ener' in arrlab:
                en_units = 'eV'
                has_energy = True

        # print("C : ", has_energy, self.workgroup.datatype, config)

        if self.workgroup.datatype in (None, 'unknown'):
            self.workgroup.datatype = 'xas' if has_energy else 'xydata'

        en_units = 'eV' if self.workgroup.datatype == 'xas' else 'unknown'

        self.read_ok_cb = read_ok_cb
        self.config = dict(xarr=None, yarr1=None, yarr2=None, yop='/',
                           ypop='', monod=3.1355316, en_units=en_units,
                           yerr_op='constant', yerr_val=1, yerr_arr=None,
                           yrpop='', yrop='/', yref1='', yref2='',
                           has_yref=False, dtc_config={}, multicol_config={})
        if config is not None:
            if 'datatype' in config:
                config.pop('datatype')
            self.config.update(config)

        if self.config['yarr2'] is None and 'i0' in self.array_labels:
            self.config['yarr2'] = 'i0'

        if self.config['yarr1'] is None:
            if 'itrans' in self.array_labels:
                self.config['yarr1'] = 'itrans'
            elif 'i1' in self.array_labels:
                self.config['yarr1'] = 'i1'

        if self.config['yref1'] is None:
            if 'iref' in self.array_labels:
                self.config['yref1'] = 'iref'
            elif 'irefer' in self.array_labels:
                self.config['yref1'] = 'irefer'
            elif 'i2' in self.array_labels:
                self.config['yref1'] = 'i2'

        if self.config['yref2'] is None and 'i1' in self.array_labels:
            self.config['yref2'] = 'i1'

        use_trans = self.config.get('is_trans', False) or 'log' in self.config['ypop']

        message = "Data Columns for %s" % group.filename
        wx.Frame.__init__(self, None, -1,
                          'Build Arrays from Data Columns for %s' % group.filename,
                          style=FRAMESTYLE)

        x0, y0 = parent.GetPosition()
        self.SetPosition((x0+60, y0+60))

        self.SetFont(Font(FONTSIZE))
        panel = wx.Panel(self)
        self.SetMinSize((725, 700))
        self.colors = GUIColors()

        def subtitle(s, fontsize=12, colour=wx.Colour(10, 10, 180)):
            return SimpleText(panel, s, font=Font(fontsize),
                           colour=colour, style=LEFT)

        # title row
        title = subtitle(message, colour=self.colors.title)

        yarr_labels = self.yarr_labels = self.array_labels + ['1.0', '']
        xarr_labels = self.xarr_labels = self.array_labels + ['_index']

        self.xarr   = Choice(panel, choices=xarr_labels, action=self.onXSelect, size=(150, -1))
        self.yarr1  = Choice(panel, choices= self.array_labels, action=self.onUpdate, size=(150, -1))
        self.yarr2  = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yerr_arr = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yerr_arr.Disable()

        self.datatype = Choice(panel, choices=DATATYPES, action=self.onUpdate, size=(150, -1))
        self.datatype.SetStringSelection(self.workgroup.datatype)

        self.en_units = Choice(panel, choices=ENUNITS_TYPES,
                               action=self.onEnUnitsSelect, size=(150, -1))

        self.ypop = Choice(panel, choices=YPRE_OPS, action=self.onUpdate, size=(100, -1))
        self.yop =  Choice(panel, choices=ARR_OPS, action=self.onUpdate, size=(100, -1))
        self.yerr_op = Choice(panel, choices=YERR_OPS, action=self.onYerrChoice, size=(100, -1))
        self.yerr_op.SetSelection(0)

        self.is_trans = Check(panel, label='is transmission data?',
                              default=use_trans, action=self.onTransCheck)

        self.yerr_val = FloatCtrl(panel, value=1, precision=4, size=(75, -1))
        self.monod_val  = FloatCtrl(panel, value=3.1355316, precision=7, size=(75, -1))

        xlab = SimpleText(panel, ' X array = ')
        ylab = SimpleText(panel, ' Y array = ')
        units_lab = SimpleText(panel, '  Units of X array:  ')
        yerr_lab = SimpleText(panel, ' Y uncertainty = ')
        dtype_lab = SimpleText(panel, ' Data Type: ')
        monod_lab = SimpleText(panel, ' Mono D spacing (Ang): ')
        yerrval_lab = SimpleText(panel, ' Value:')
        self.info_message = subtitle('    ', colour=wx.Colour(100, 10, 10))

        # yref
        self.has_yref = Check(panel, label='data file includes energy reference data',
                              default=self.config['has_yref'],
                              action=self.onYrefCheck)
        refylab = SimpleText(panel, ' Refer array = ')
        self.yref1 = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yref2 = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yrpop = Choice(panel, choices=YPRE_OPS, action=self.onUpdate, size=(100, -1))
        self.yrop =  Choice(panel, choices=ARR_OPS, action=self.onUpdate, size=(100, -1))

        self.ysuf = SimpleText(panel, '')
        # print("COL FILE READER set ypop to ", use_trans, self.config['ypop'])
        self.ypop.SetStringSelection(self.config['ypop'])
        self.yop.SetStringSelection(self.config['yop'])
        self.yrpop.SetStringSelection(self.config['yrpop'])
        self.yrop.SetStringSelection(self.config['yrop'])
        self.monod_val.SetValue(self.config['monod'])
        self.monod_val.SetAction(self.onUpdate)
        self.monod_val.Enable(self.config['en_units'].startswith('deg'))
        self.en_units.SetStringSelection(self.config['en_units'])
        self.yerr_op.SetStringSelection(self.config['yerr_op'])
        self.yerr_val.SetValue(self.config['yerr_val'])
        if '(' in self.config['ypop']:
            self.ysuf.SetLabel(')')


        ixsel, iysel = 0, 1
        iy2sel = iyesel = iyr1sel = iyr2sel = len(yarr_labels)-1
        if self.config['xarr'] in xarr_labels:
            ixsel = xarr_labels.index(self.config['xarr'])
        if self.config['yarr1'] in self.array_labels:
            iysel = self.array_labels.index(self.config['yarr1'])
        if self.config['yarr2'] in yarr_labels:
            iy2sel = yarr_labels.index(self.config['yarr2'])
        if self.config['yerr_arr'] in yarr_labels:
            iyesel = yarr_labels.index(self.config['yerr_arr'])
        if self.config['yref1'] in self.array_labels:
            iyr1sel = self.array_labels.index(self.config['yref1'])
        if self.config['yref2'] in self.array_labels:
            iyr2sel = self.array_labels.index(self.config['yref2'])

        self.xarr.SetSelection(ixsel)
        self.yarr1.SetSelection(iysel)
        self.yarr2.SetSelection(iy2sel)
        self.yerr_arr.SetSelection(iyesel)
        self.yref1.SetSelection(iyr1sel)
        self.yref2.SetSelection(iyr2sel)

        self.wid_filename = wx.TextCtrl(panel, value=fix_filename(group.filename),
                                         size=(250, -1))
        self.wid_groupname = wx.TextCtrl(panel, value=group.groupname,
                                         size=(150, -1))
        if not edit_groupname:
            self.wid_groupname.Disable()
        self.wid_reffilename = wx.TextCtrl(panel, value=fix_filename(group.filename + '_ref'),
                                         size=(250, -1))
        self.wid_refgroupname = wx.TextCtrl(panel, value=group.groupname + '_ref',
                                         size=(150, -1))

        self.onTransCheck(is_trans=use_trans)
        self.onYrefCheck(has_yref=self.config['has_yref'])


        bpanel = wx.Panel(panel)
        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        _ok    = Button(bpanel, 'OK', action=self.onOK)
        _cancel = Button(bpanel, 'Cancel', action=self.onCancel)
        _edit   = Button(bpanel, 'Edit Array Names', action=self.onEditNames)
        self.multi_sel = Button(bpanel, 'Select Multilple Columns',  action=self.onMultiColumn)
        self.multi_clear = Button(bpanel, 'Clear Multiple Columns',  action=self.onClearMultiColumn)
        self.multi_clear.Disable()
        _edit.SetToolTip('Change the current Column Names')

        self.multi_sel.SetToolTip('Select Multiple Columns to import as separate groups')
        self.multi_clear.SetToolTip('Clear Multiple Column Selection')
        bsizer.Add(_ok)
        bsizer.Add(_cancel)
        bsizer.Add(_edit)
        bsizer.Add(self.multi_sel)
        bsizer.Add(self.multi_clear)
        _ok.SetDefault()
        pack(bpanel, bsizer)

        self.dtc_button  = Button(panel, 'Sum and Correct Fluoresence Data', action=self.onDTC)
        self.dtc_button.SetToolTip('Select channels and do deadtime-corrections for multi-element fluorescence data')

        sizer = wx.GridBagSizer(2, 2)
        sizer.Add(title,     (0, 0), (1, 7), LEFT, 5)

        ir = 1
        sizer.Add(subtitle(' X [Energy] Array:'),   (ir, 0), (1, 2), LEFT, 0)
        sizer.Add(dtype_lab,       (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.datatype,   (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(xlab,      (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.xarr, (ir, 1), (1, 2), LEFT, 0)
        sizer.Add(units_lab,     (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.en_units,  (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(monod_lab,     (ir, 2), (1, 2), RIGHT, 0)
        sizer.Add(self.monod_val,(ir, 4), (1, 1), LEFT, 0)

        ir += 1
        sizer.Add(subtitle(' Y [\u03BC(E)] Array:'), (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.is_trans,                     (ir, 1), (1, 2), LEFT, 0)
        sizer.Add(self.dtc_button,                   (ir, 3), (1, 2), RIGHT, 0)
        ir += 1
        sizer.Add(ylab,       (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.ypop,  (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(self.yarr1, (ir, 2), (1, 1), LEFT, 0)
        sizer.Add(self.yop,   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.yarr2, (ir, 4), (1, 1), LEFT, 0)
        sizer.Add(self.ysuf,  (ir, 5), (1, 1), LEFT, 0)


        ir += 1
        sizer.Add(yerr_lab,      (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.yerr_op,  (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(self.yerr_arr, (ir, 2), (1, 1), LEFT, 0)
        sizer.Add(yerrval_lab,   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.yerr_val, (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(SimpleText(panel, ' Display Name:'), (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.wid_filename,                  (ir, 1), (1, 2), LEFT, 0)
        sizer.Add(SimpleText(panel, ' Group Name:'),   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.wid_groupname,                 (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(self.info_message,                  (ir, 0), (1, 5), LEFT, 1)

        ir += 2
        sizer.Add(subtitle(' Reference [\u03BC_ref(E)] Array: '),
                  (ir, 0), (1, 2), LEFT, 0)
        sizer.Add(self.has_yref,   (ir, 2), (1, 3), LEFT, 0)

        ir += 1
        sizer.Add(refylab,    (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.yrpop, (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(self.yref1, (ir, 2), (1, 1), LEFT, 0)
        sizer.Add(self.yrop,  (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.yref2, (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(SimpleText(panel, ' Reference Name:'), (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.wid_reffilename,               (ir, 1), (1, 2), LEFT, 0)
        sizer.Add(SimpleText(panel, ' Group Name:'),   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.wid_refgroupname,              (ir, 4), (1, 2), LEFT, 0)

        ir +=1
        sizer.Add(bpanel,     (ir, 0), (1, 5), LEFT, 3)

        pack(panel, sizer)

        self.nb = fnb.FlatNotebook(self, -1, agwStyle=FNB_STYLE)
        self.nb.SetTabAreaColour(wx.Colour(248,248,240))
        self.nb.SetActiveTabColour(wx.Colour(254,254,195))
        self.nb.SetNonActiveTabTextColour(wx.Colour(40,40,180))
        self.nb.SetActiveTabTextColour(wx.Colour(80,0,0))

        self.plotpanel = PlotPanel(self, messenger=self.plot_messages)
        try:
            plotopts = self._larch.symtable._sys.wx.plotopts
            self.plotpanel.conf.set_theme(plotopts['theme'])
            self.plotpanel.conf.enable_grid(plotopts['show_grid'])
        except:
            pass


        self.plotpanel.SetMinSize((200, 200))
        textpanel = wx.Panel(self)
        ftext = wx.TextCtrl(textpanel, style=wx.TE_MULTILINE|wx.TE_READONLY,
                               size=(400, 250))

        ftext.SetValue(group.text)
        ftext.SetFont(Font(FONTSIZE))

        textsizer = wx.BoxSizer(wx.VERTICAL)
        textsizer.Add(ftext, 1, LEFT|wx.GROW, 1)
        pack(textpanel, textsizer)

        self.nb.AddPage(textpanel, ' Text of Data File ', True)
        self.nb.AddPage(self.plotpanel, ' Plot of Selected Arrays ', True)

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel, 0, wx.GROW|wx.ALL, 2)
        mainsizer.Add(self.nb, 1, LEFT|wx.GROW,   2)
        pack(self, mainsizer)

        self.statusbar = self.CreateStatusBar(2, 0)
        self.statusbar.SetStatusWidths([-1, -1])
        statusbar_fields = [group.filename, ""]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)

        self.set_energy_units()
        dtc_conf = self.config.get('dtc_config', {})
        if len(dtc_conf) > 0:
            self.onDTC_OK(dtc_conf, update=False)

        self.Show()
        self.Raise()
        self.onUpdate()

    def onDTC(self, event=None):
        self.show_subframe('dtc_conf', DeadtimeCorrectionFrame,
                           config=self.config['dtc_config'],
                           group=self.workgroup,
                           on_ok=self.onDTC_OK)

    def onDTC_OK(self, config, update=True, **kws):
        label, sum = sum_fluor_channels(self.workgroup, config['roi'],
                                        icr=config['icr'],
                                        ocr=config['ocr'],
                                        ltime=config['ltime'],
                                        add_data=False)
        if sum is None:
            return
        self.info_message.SetLabel(f"Added array '{label}' with summed and corrected fluorecence data")
        self.workgroup.array_labels.append(label)
        self.set_array_labels(self.workgroup.array_labels)
        npts = len(sum)
        new = np.append(self.workgroup.raw.data, sum.reshape(1, npts), axis=0)
        self.workgroup.raw.data = new[()]
        self.workgroup.data = new[()]
        self.yarr1.SetStringSelection(label)
        self.config['dtc_config'] = config
        if update:
            self.onUpdate()

    def onClearMultiColumn(self, event=None):
        self.config['multicol_config'] = {}
        self.info_message.SetLabel(f" cleared reading of multiple columns")
        self.multi_clear.Disable()
        self.yarr1.Enable()
        self.ypop.Enable()
        self.yop.Enable()
        self.onUpdate()


    def onMultiColumn(self, event=None):
        self.show_subframe('multicol', MultiColumnFrame,
                           config=self.config['multicol_config'],
                           group=self.workgroup,
                           on_ok=self.onMultiColumn_OK)


    def onMultiColumn_OK(self, config, update=True, **kws):
        chans = config.get('channels', [])
        if len(chans) == 0:
            self.config['multicol_config'] = {}
        else:
            self.config['multicol_config'] = config
            self.yarr1.SetSelection(chans[0])
            self.yarr2.SetSelection(config['i0'])
            self.ypop.SetStringSelection('')
            self.yarr1.Disable()
            self.ypop.Disable()
            self.yop.Disable()
            y2 = self.yarr2.GetStringSelection()
            msg = f"  Will import {len(config['channels'])} Y arrays, divided by '{y2}'"
            self.info_message.SetLabel(msg)
            self.multi_clear.Enable()
        if update:
            self.onUpdate()


    def read_column_file(self, path):
        """read column file, generally as initial read"""
        path = Path(path).absolute()
        filename = path.name
        path = path.as_posix()
        reader, text = guess_filereader(path, return_text=True)

        if reader == 'read_specfile':
            if not is_specfile(path, require_multiple_scans=True):
                reader = 'read_ascii'

        if reader in ('read_xdi', 'read_gsexdi'):
            # first check for Nans and Infs
            nan_result = look_for_nans(path)
            if 'read error' in nan_result.message:
                title = "Cannot read %s" % path
                message = "Error reading %s\n%s" %(path, nan_result.message)
                r = Popup(self.parent, message, title)
                return None
            if 'no data' in nan_result.message:
                title = "No data in %s" % path
                message = "No data found in file %s" % path
                r = Popup(self.parent, message, title)
                return None

            if ('has nans' in nan_result.message or
                'has infs' in nan_result.message):
                reader = 'read_ascii'

        tmpname = '_tmpfile_'
        read_cmd = "%s = %s('%s')" % (tmpname, reader, path)
        self.reader = reader
        _larch = self._larch

        if (not isinstance(_larch, larch.Interpreter) and
            hasattr(_larch, '_larch')):
            _larch = _larch._larch
        try:
            _larch.eval(read_cmd, add_history=True)
        except:
            pass
        if len(_larch.error) > 0 and reader in ('read_xdi', 'read_gsexdi'):
            read_cmd = "%s = %s('%s')" % (tmpname, 'read_ascii', path)
            try:
                _larch.eval(read_cmd, add_history=True)
            except:
                pass
            if len(_larch.error) == 0:
                self.reader = 'read_ascii'

        if len(_larch.error) > 0:
            msg = ["Error trying to read '%s':" % path, ""]
            for err in _larch.error:
                exc_name, errmsg = err.get_error()
                msg.append(errmsg)

            title = "Cannot read %s" % path
            r = Popup(self.parent, "\n".join(msg), title)
            return None
        group = deepcopy(_larch.symtable.get_symbol(tmpname))
        _larch.symtable.del_symbol(tmpname)

        group.text = text
        group.path = path
        group.filename = filename
        group.groupname = file2groupname(filename,
                                         symtable=self._larch.symtable)
        return group

    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                pass
        if not shown:
            self.subframes[name] = frameclass(self, **opts)
            self.subframes[name].Show()
            self.subframes[name].Raise()


    def onEditNames(self, evt=None):
        self.show_subframe('editcol', EditColumnFrame,
                           group=self.workgroup,
                           on_ok=self.set_array_labels)

    def set_array_labels(self, arr_labels):
        self.workgroup.array_labels = arr_labels
        yarr_labels = self.yarr_labels = arr_labels + ['1.0', '']
        xarr_labels = self.xarr_labels = arr_labels + ['_index']
        def update(wid, choices):
            curstr = wid.GetStringSelection()
            curind = wid.GetSelection()
            wid.SetChoices(choices)
            if curstr in choices:
                wid.SetStringSelection(curstr)
            else:
                wid.SetSelection(curind)
        update(self.xarr,  xarr_labels)
        update(self.yarr1, yarr_labels)
        update(self.yarr2, yarr_labels)
        update(self.yerr_arr, yarr_labels)
        self.onUpdate()

    def onOK(self, event=None):
        """ build arrays according to selection """
        self.read_form()
        cout = create_arrays(self.workgroup, **self.config)
        self.config.update(cout)
        conf = self.config
        if self.ypop.Enabled:  #not using multicolumn mode
            conf['multicol_config'] = {'channels': [], 'i0': conf['iy2']}

        self.expressions = conf['expressions']
        filename = conf['filename']
        groupname = conf['groupname']

        conf['array_labels'] = self.workgroup.array_labels

        # generate script to pass back to calling program:
        labstr = ', '.join(self.array_labels)
        buff = [f"{{group}} = {self.reader}('{{path}}', labels='{labstr}')",
                "{group}.path = '{path}'",
                "{group}.is_frozen = False",
                "{group}.energy_ref = '{group}'"]

        dtc_conf = conf.get('dtc_config', {})
        if len(dtc_conf) > 0:
            sumcmd = "sum_fluor_channels({{group}}, {roi}, icr={icr}, ocr={ocr}, ltime={ltime})"
            buff.append(sumcmd.format(**dtc_conf))

        buff.append("{group}.datatype = '%s'" % (conf['datatype']))

        for attr in ('plot_xlabel', 'plot_ylabel'):
            val = getattr(self.workgroup, attr)
            buff.append("{group}.%s = '%s'" % (attr, val))

        xexpr = self.expressions['xplot']
        en_units = conf['en_units']
        if en_units.startswith('deg'):
            monod = conf['monod']
            buff.append(f"monod = {monod:.9f}")
            buff.append(f"{{group}}.xplot = PLANCK_HC/(2*monod*sin(DEG2RAD*({xexpr:s})))")
        elif en_units.startswith('keV'):
            buff.append(f"{{group}}.xplot = 1000.0*{xexpr:s}")
        else:
            buff.append(f"{{group}}.xplot = {xexpr:s}")

        for aname in ('yplot', 'yerr'):
            expr = self.expressions[aname]
            buff.append(f"{{group}}.{aname} = {expr}")


        dtype = getattr(self.workgroup, 'datatype', 'xytype')
        if dtype == 'xas':
            if self.reader == 'read_gsescan':
                buff.append("{group}.xplot = {group}.x")
            buff.append("{group}.energy = {group}.xplot[:]")
            buff.append("{group}.mu = {group}.yplot[:]")
            buff.append("sort_xafs({group}, overwrite=True, fix_repeats=True)")
        elif dtype == 'xydata':
            buff.append("{group}.x = {group}.xplot[:]")
            buff.append("{group}.y = {group}.yplot[:]")
            buff.append("{group}.scale = (ptp({group}.yplot)+1.e-15)")
            buff.append("{group}.xshift = 0.0")

        array_desc = dict(xplot=self.workgroup.plot_xlabel,
                          yplot=self.workgroup.plot_ylabel,
                          yerr=self.expressions['yerr'])

        reffile = refgroup = None
        if conf['has_yref']:
            reffile = conf['reffile']
            refgroup = conf['refgroup']
            refexpr = self.expressions['yref']
            array_desc['yref'] = getattr(self.workgroup, 'yrlabel', 'reference')

            buff.append("# reference group")
            buff.append("{refgroup} = deepcopy({group})")
            buff.append(f"{{refgroup}}.yplot = {{refgroup}}.mu = {refexpr}")
            buff.append(f"{{refgroup}}.plot_ylabel = '{self.workgroup.yrlabel}'")
            buff.append("{refgroup}.energy_ref = {group}.energy_ref = '{refgroup}'")
            buff.append("# end reference group")

        script = "\n".join(buff)
        conf['array_desc'] = array_desc

        if self.read_ok_cb is not None:
            self.read_ok_cb(script, self.path, conf)

        for f in self.subframes.values():
            try:
                f.Destroy()
            except:
                pass
        self.Destroy()

    def onCancel(self, event=None):
        self.workgroup.import_ok = False
        for f in self.subframes.values():
            try:
                f.Destroy()
            except:
                pass
        self.Destroy()

    def onYerrChoice(self, evt=None):
        yerr_choice = evt.GetString()
        self.yerr_arr.Disable()
        self.yerr_val.Disable()
        if 'const' in yerr_choice.lower():
            self.yerr_val.Enable()
        elif 'array' in yerr_choice.lower():
            self.yerr_arr.Enable()
        # self.onUpdate()

    def onTransCheck(self, evt=None, is_trans=False):
        if evt is not None:
            is_trans = evt.IsChecked()
        if is_trans:
            self.ypop.SetStringSelection('-log(')
        else:
            self.ypop.SetStringSelection('')
        try:
            self.onUpdate()
        except:
            pass

    def onYrefCheck(self, evt=None, has_yref=False):
        if evt is not None:
            has_yref = evt.IsChecked()
        self.yref1.Enable(has_yref)
        self.yref2.Enable(has_yref)
        self.yrpop.Enable(has_yref)
        self.yrop.Enable(has_yref)
        self.wid_reffilename.Enable(has_yref)
        self.wid_refgroupname.Enable(has_yref)


    def onXSelect(self, evt=None):
        ix  = self.xarr.GetSelection()
        xname = self.xarr.GetStringSelection()

        workgroup = self.workgroup
        ncol, npts = self.workgroup.data.shape
        if xname.startswith('_index') or ix >= ncol:
            workgroup.xplot = 1.0*np.arange(npts)
        else:
            workgroup.xplot = 1.0*self.workgroup.data[ix, :]
        self.onUpdate()

        self.monod_val.Disable()
        if self.datatype.GetStringSelection().strip().lower() == 'xydata':
            self.en_units.SetSelection(4)
        else:
            eguess = guess_energy_units(workgroup.xplot)
            if eguess.startswith('keV'):
                self.en_units.SetSelection(1)
            elif eguess.startswith('deg'):
                self.en_units.SetSelection(2)
                self.monod_val.Enable()
            else:
                self.en_units.SetSelection(0)

    def onEnUnitsSelect(self, evt=None):
        self.monod_val.Enable(self.en_units.GetStringSelection().startswith('deg'))
        self.onUpdate()

    def set_energy_units(self):
        ix  = self.xarr.GetSelection()
        xname = self.xarr.GetStringSelection()
        workgroup = self.workgroup
        try:
            ncol, npts = workgroup.data.shape
        except (AttributeError,  ValueError):
            return

        if xname.startswith('_index') or ix >= ncol:
            workgroup.xplot = 1.0*np.arange(npts)
        else:
            workgroup.xplot = 1.0*self.workgroup.data[ix, :]
        if self.datatype.GetStringSelection().strip().lower() != 'xydata':
            eguess =  guess_energy_units(workgroup.xplot)
            if eguess.startswith('eV'):
                self.en_units.SetStringSelection('eV')
            elif eguess.startswith('keV'):
                self.en_units.SetStringSelection('keV')

    def read_form(self, **kws):
        """return form configuration"""
        datatype = self.datatype.GetStringSelection().strip().lower()
        if self.workgroup.datatype == 'xydata' and datatype == 'xas':
            self.workgroup.datatype = 'xas'
            eguess = guess_energy_units(self.workgroup.xplot)
            if eguess.startswith('keV'):
                self.en_units.SetSelection(1)
            elif eguess.startswith('deg'):
                self.en_units.SetSelection(2)
                self.monod_val.Enable()
            else:
                self.en_units.SetSelection(0)
        if datatype == 'xydata':
            self.en_units.SetStringSelection('not energy')

        ypop = self.ypop.GetStringSelection().strip()
        self.is_trans.SetValue('log' in ypop)


        conf = {'datatype': datatype,
                'ix':  self.xarr.GetSelection(),
                'xarr': self.xarr.GetStringSelection(),
                'en_units': self.en_units.GetStringSelection(),
                'monod': float(self.monod_val.GetValue()),
                'yarr1': self.yarr1.GetStringSelection().strip(),
                'yarr2': self.yarr2.GetStringSelection().strip(),
                'iy1': self.yarr1.GetSelection(),
                'iy2': self.yarr2.GetSelection(),
                'yop': self.yop.GetStringSelection().strip(),
                'ypop': ypop,
                'iyerr': self.yerr_arr.GetSelection(),
                'yerr_arr': self.yerr_arr.GetStringSelection(),
                'yerr_op': self.yerr_op.GetStringSelection().lower(),
                'yerr_val': self.yerr_val.GetValue(),
                'has_yref': self.has_yref.IsChecked(),
                'yref1': self.yref1.GetStringSelection().strip(),
                'yref2': self.yref2.GetStringSelection().strip(),
                'iry1': self.yref1.GetSelection(),
                'iry2': self.yref2.GetSelection(),
                'yrpop': self.yrpop.GetStringSelection().strip(),
                'yrop': self.yop.GetStringSelection().strip(),
                'filename': self.wid_filename.GetValue(),
                'groupname': fix_varname(self.wid_groupname.GetValue()),
                'reffile': self.wid_reffilename.GetValue(),
                'refgroup': fix_varname(self.wid_refgroupname.GetValue()),
                }
        self.config.update(conf)
        return conf

    def onUpdate(self, evt=None, **kws):
        """column selections changed calc xplot and yplot"""
        workgroup = self.workgroup
        try:
            ncol, npts = self.workgroup.data.shape
        except:
            return

        conf = self.read_form()
        cout = create_arrays(workgroup, **conf)
        self.expressions = cout.pop('expressions')
        conf.update(cout)

        if energy_may_need_rebinning(workgroup):
            self.info_message.SetLabel("Warning: XAS data may need to be rebinned!")

        fname = Path(workgroup.filename).name
        popts = dict(marker='o', markersize=4, linewidth=1.5, title=fname,
                     xlabel=workgroup.plot_xlabel,
                     ylabel=workgroup.plot_ylabel,
                     label=workgroup.plot_ylabel)

        self.plotpanel.plot(workgroup.xplot, workgroup.yplot, **popts)
        if conf['has_yref']:
            yrlabel = getattr(workgroup, 'plot_yrlabel', 'reference')
            self.plotpanel.oplot(workgroup.xplot, workgroup.yref,
                                 y2label=yrlabel,
                                 linewidth=2.0, color='#E08070',
                                 label=yrlabel, zorder=-40, side='right')

        for i in range(self.nb.GetPageCount()):
            if 'plot' in self.nb.GetPageText(i).lower():
                self.nb.SetSelection(i)

    def plot_messages(self, msg, panel=1):
        self.statusbar.SetStatusText(msg, panel)


def create_arrays(dgroup, datatype='xas', ix=0, xarr='energy', en_units='eV',
                  monod=3.1355316, yarr1=None, yarr2=None, iy1=2, iy2=1, yop='/',
                  ypop='', iyerr=5, yerr_arr=None, yerr_op='constant', yerr_val=1.0,
                  has_yref=False, yref1=None, yref2=None, iry1=3, iry2=2,
                  yrpop='', yrop='/', **kws):
    """
    build arrays and values for datagroup based on configuration as from ColumnFile
    """
    ncol, npts = dgroup.data.shape
    exprs = dict(xplot=None, yplot=None, yerr=None, yref=None)

    if not hasattr(dgroup, 'index'):
        dgroup.index = 1.0*np.arange(npts)

    if xarr.startswith('_index') or ix >= ncol:
        dgroup.xplot = 1.0*np.arange(npts)
        xarr = '_index'
        exprs['xplot'] = 'arange({npts})'
    else:
        dgroup.xplot = 1.0*dgroup.data[ix, :]
        exprs['xplot'] = '{group}.data[{ix}, : ]'

    xlabel = xarr
    monod = float(monod)
    if en_units.startswith('deg'):
        dgroup.xplot = PLANCK_HC/(2*monod*np.sin(DEG2RAD*dgroup.xplot))
        xlabel = xarr + ' (eV)'
    elif en_units.startswith('keV'):
        dgroup.xplot *= 1000.0
        xlabel = xarr + ' (eV)'

    def pre_op(opstr, arr):
        if opstr == '-':
            return '', opstr, -arr
        suf = ''
        if opstr in ('-log(', 'log('):
            suf = ')'
            arr = safe_log(arr)
            if opstr.startswith('-'): arr = -arr
            arr[np.where(np.isnan(arr))] = 0
        return suf, opstr, arr

    if yarr1 is None:
        yarr1 = dgroup.array_labels[iy1]

    if yarr2 is None:
        yarr2 = dgroup.array_labels[iy2]

    ylabel = yarr1
    if len(yarr2) == 0:
        yarr2 = '1.0'
    else:
        ylabel = f"{ylabel}{yop}{yarr2}"

    if yarr1 == '0.0':
        ydarr1 = np.zeros(npts)*1.0
        yexpr1 = f'np.zeros(npts)'
    elif len(yarr1) == 0 or yarr1 == '1.0' or iy1 >= ncol:
        ydarr1 = np.ones(npts)*1.0
        yexpr1 = f'np.ones({npts})'
    else:
        ydarr1 = dgroup.data[iy1, :]
        yexpr1 = '{group}.data[{iy1}, : ]'

    dgroup.yplot = ydarr1
    exprs['yplot'] = yexpr1

    if yarr2 == '0.0':
        ydarr2 = np.zeros(npts)*1.0
        yexpr2 = '0.0'
    elif len(yarr2) == 0 or yarr2 == '1.0' or iy2 >= ncol:
        ydarr2 = np.ones(npts)*1.0
        yexpr2 = '1.0'
    else:
        ydarr2 = dgroup.data[iy2, :]
        yexpr2 = '{group}.data[{iy2}, : ]'

    if yop in ('+', '-', '*', '/'):
        exprs['yplot'] = f"{yexpr1}{yop}{yexpr2}"
        if yop == '+':
            dgroup.yplot = ydarr1 + ydarr2
        elif yop == '-':
            dgroup.yplot = ydarr1 - ydarr2
        elif yop == '*':
            dgroup.yplot = ydarr1 * ydarr2
        elif yop == '/':
            dgroup.yplot = ydarr1 / ydarr2

    ysuf, ypop, dgroup.yplot = pre_op(ypop, dgroup.yplot)
    ypopx = ypop.replace('log', 'safe_log')
    exprs['yplot'] = f"{ypopx}{exprs['yplot']}{ysuf}"
    ylabel = f"{ypop}{ylabel}{ysuf}"

    # error
    exprs['yerr'] = '1'
    if yerr_op.startswith('const'):
        yderr = yerr_val
        exprs['yerr'] = f"{yerr_val}"
    elif yerr_op.startswith('array'):
        yderr = dgroup.data[iyerr, :]
        exprs['yerr'] = '{group}.data[{iyerr}, :]'
    elif yerr_op.startswith('sqrt'):
        yderr = np.sqrt(dgroup.yplot)
        exprs['yerr'] = 'sqrt({group}.yplot)'

    # reference
    yrlabel = None
    if has_yref:
        yrlabel = yref1
        if len(yref2) == 0:
            yref2 = '1.0'
        else:
            yrlabel = f"{yrlabel}{yrop}{yref2}"

        if yref1 == '0.0':
            ydrarr1 = np.zeros(npts)*1.0
            yrexpr1 = 'zeros({npts})'
        elif len(yref1) == 0 or yref1 == '1.0' or iry1 >= ncol:
            ydrarr1 = np.ones(npts)*1.0
            yrexpr1 = 'ones({npts})'
        else:
            ydrarr1 = dgroup.data[iry1, :]
            yrexpr1 = '{group}.data[{iry1}, : ]'

        dgroup.yref = ydrarr1
        exprs['yref'] = yrexpr1

        if yref2 == '0.0':
            ydrarr2 = np.zeros(npts)*1.0
            ydrexpr2 = '0.0'
        elif len(yref2) == 0 or yref2 == '1.0' or iry2 >= ncol:
            ydrarr2 = np.ones(npts)*1.0
            yrexpr2 = '1.0'
        else:
            ydrarr2 = dgroup.data[iry2, :]
            yrexpr2 = '{group}.data[{iry2}, : ]'

        if yrop in ('+', '-', '*', '/'):
            exprs['yref'] = f'{yrexpr1} {yop} {yrexpr2}'
            if yrop == '+':
                dgroup.yref = ydrarr1 + ydrarr2
            elif yrop == '-':
                dgroup.yref = ydrarr1 - ydrarr2
            elif yrop == '*':
                dgroup.yref = ydrarr1 * ydarr2
            elif yrop == '/':
                dgroup.yref = ydrarr1 / ydrarr2

        yrsuf, yprop, dgroup.yref = pre_op(yrpop, dgroup.yref)
        yrpopx = yrpop.replace('log', 'safe_log')
        exprs['yref'] = f"{yrpopx}{exprs['yref']}{yrsuf}"
        yrlabel = f'{yrpop} {yrlabel} {yrsuf}'
        dgroup.yrlabel = yrlabel


    try:
        npts = min(len(dgroup.xplot), len(dgroup.yplot))
    except AttributeError:
        return
    except ValueError:
        return

    en = dgroup.xplot
    dgroup.datatype    = datatype
    dgroup.npts        = npts
    dgroup.plot_xlabel = xlabel
    dgroup.plot_ylabel = ylabel
    dgroup.xplot       = np.array(dgroup.xplot[:npts])
    dgroup.yplot       = np.array(dgroup.yplot[:npts])
    dgroup.y           = dgroup.yplot
    dgroup.yerr        = yderr
    if isinstance(yderr, np.ndarray):
        dgroup.yerr    = np.array(yderr[:npts])
    if yrlabel is not None:
        dgroup.plot_yrlabel = yrlabel

    if dgroup.datatype == 'xas':
        dgroup.energy = dgroup.xplot
        dgroup.mu     = dgroup.yplot

    return dict(xarr=xarr, ypop=ypop, yop=yop, yarr1=yarr1, yarr2=yarr2,
                monod=monod, en_units=en_units, yerr_op=yerr_op,
                yerr_val=yerr_val, yerr_arr=yerr_arr, yrpop=yrpop, yrop=yrop,
                yref1=yref1, yref2=yref2, has_yref=has_yref,
                expressions=exprs)

def energy_may_need_rebinning(workgroup):
    "test if energy may need rebinning"
    if getattr(workgroup, 'datatype', '?') != 'xas':
        return False
    en = getattr(workgroup, 'xplot', [-8.0e12])
    if len(en) < 2:
        return False
    if not isinstance(en, np.ndarray):
        en = np.array(en)
    if len(en) > 2000 or any(np.diff(en))< 0:
        return True
    if (len(en) > 200 and (max(en) - min(en)) > 350 and
        np.diff(en[:-100]).mean() < 1.0):
        return True
